"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as ppa
import pyparsing.common as ppc

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_parsing.tokens.base_tokens import BaseTokenWithPos
from gidapptools.gid_parsing.universal.datetime_elements import get_grammar_from_dt_format
from gidapptools.gid_parsing.universal.character_elements import Ligatures, BaseElements

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class LineNumberToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: int) -> None:
        super().__init__(start, end)
        self.line_number = value


class LogLevelToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: str) -> None:
        super().__init__(start, end)
        self.log_level = value


class ThreadNameToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: str) -> None:
        super().__init__(start, end)
        self.thread_name = value


class ModuleNameToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: str) -> None:
        super().__init__(start, end)
        self.module_name = value
        self.module_name_parts = tuple(self.module_name.split('.'))

    @property
    def parent_module_name(self) -> str:
        try:
            return self.module_name_parts[-2]
        except IndexError:
            return "__main__"


class FunctionNameToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: str) -> None:
        super().__init__(start, end)
        self.function_name = value


class MessageToken(BaseTokenWithPos):

    def __init__(self, start: int, end: int, value: str) -> None:
        super().__init__(start, end)
        self.message = value


class GeneralGrammar:
    _cached_grammar: ppa.ParserElement = None

    @property
    def ___grammar___(self) -> ppa.ParserElement:
        if self._cached_grammar is None:
            self.__class__._cached_grammar = self._construct_grammar()
        return self._cached_grammar

    def _construct_grammar(self):
        non_pipe_printables = ppa.printables.replace("|", "")
        time_stamp_part = get_grammar_from_dt_format("%Y-%m-%d %H:%M:%S.%f %Z")("time_stamp")
        line_number_part = ppa.locatedExpr(ppa.Word(ppa.nums).set_parse_action(ppc.convert_to_integer)).set_parse_action(LineNumberToken.from_parse_action)("line_number")
        level_part = ppa.locatedExpr(ppa.Keyword("DEBUG") | ppa.Keyword("INFO") | ppa.Keyword("CRITICAL") | ppa.Keyword("ERROR")).set_parse_action(LogLevelToken.from_parse_action)("level")
        thread_name_part = ppa.locatedExpr(ppa.Word(non_pipe_printables)("thread")).set_parse_action(ThreadNameToken.from_parse_action)("thread")
        module_name_part = ppa.locatedExpr(ppa.Word(non_pipe_printables)).set_parse_action(ModuleNameToken.from_parse_action)("module")
        function_name_part = ppa.locatedExpr(ppa.Word(non_pipe_printables)).set_parse_action(FunctionNameToken.from_parse_action)("function")
        return time_stamp_part + BaseElements.pipe + line_number_part + BaseElements.pipe + level_part + BaseElements.pipe + thread_name_part + BaseElements.pipe + \
            module_name_part + BaseElements.pipe + function_name_part + BaseElements.pipe + BaseElements.pipe + Ligatures.big_arrow_right + ppa.locatedExpr(ppa.rest_of_line).set_parse_action(MessageToken.from_parse_action)("message")

    def __call__(self, text: str) -> dict[str, BaseTokenWithPos]:
        return self.___grammar___.parse_string(text, parse_all=True).as_dict()


# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from time import sleep, process_time_ns
from pathlib import Path
from contextlib import contextmanager

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_warning.experimental import mark_experimental

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from abc import ABCMeta
from typing import Any
from pathlib import Path
from weakref import ProxyType, proxy

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as pp

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class Token:
    __slots__ = tuple()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(" + ', '.join(f"{n}={getattr(self, n)!r}" for n in self.__slots__) + ')'


class IniToken(Token, metaclass=ABCMeta):
    __slots__ = tuple()
    spec_data = None

    def add_comment(self, comment: "Comment") -> None:
        self.comments.append(comment.content)


class Comment(Token):
    __slots__ = ("content", "comment_indicator")

    def __init__(self, content: str) -> None:
        self.content = content.strip()
        self.comment_indicator: str = '#'

    def __str__(self) -> str:
        return self.content

    def as_text(self) -> str:
        return f"{self.comment_indicator} {self}"


class Section(IniToken):
    __slots__ = ("name", "comments", "entries", "__weakref__")

    def __init__(self, name: str) -> None:
        self.name = name
        self.comments = []
        self.entries = {}

    def has_key(self, key_name: str) -> bool:
        return key_name in self.entries

    def add_entry(self, entry: "Entry") -> None:
        entry.section = proxy(self)
        self.entries[entry.key] = entry

    def remove_entry(self, entry_key: str) -> None:
        del self.entries[entry_key]

    def __getitem__(self, key: str) -> "Entry":
        return self.entries[key]

    def get(self, key, default=None) -> Any:
        try:
            return self[key].get_value()
        except KeyError:
            return default

    def __len__(self) -> int:
        return len(self.entries)

    def as_dict(self) -> dict[str, dict[str, str]]:
        data = {self.name: {}}
        for entry in self.entries.values():
            data[self.name] |= entry.as_dict()
        return data

    def as_text(self,
                section_header_newlines: int = 1,
                extra_section_newlines: int = 2,
                extra_entry_newlines: int = 0) -> str:
        lines = []
        lines += [comment.as_text() for comment in self.comments]
        lines.append(f"[{self.name}]")
        lines += ['' for i in range(section_header_newlines)]
        lines += [entry.as_text(extra_entry_newlines=extra_entry_newlines) for entry in self.entries.values()]
        lines += ['' for i in range(extra_section_newlines)]
        return '\n'.join(lines)


class EnvSection(Section):
    __slots__ = tuple()
    # pylint: disable=super-init-not-called

    def __init__(self) -> None:
        self.name = "__ENV__"
        self.comments = None

    @property
    def entries(self) -> dict[str, "Entry"]:
        return {key: Entry(key, value) for key, value in os.environ.items()}


class Entry(IniToken):
    __slots__ = ("key", "value", "key_value_separator", "comments", "section")

    def __init__(self, key: str, value: str = None) -> None:
        self.key = key.strip()
        self.value = value.lstrip() if value is not None else value
        self.key_value_separator = '='
        self.comments = []
        self.section: ProxyType[Section] = None

    def as_dict(self) -> dict[str, str]:
        return {self.key: self.value}

    def as_text(self, extra_entry_newlines: int = 0) -> str:
        lines = [comment.as_text() for comment in self.comments]
        text_value = "" if self.value is None else self.value
        lines.append(f"{self.key} {self.key_value_separator} {text_value}")
        lines += ['' for i in range(extra_entry_newlines)]
        return '\n'.join(lines)


class TokenFactory:
    __slots__ = ("token_map",)

    def __init__(self, token_map: dict[str, type] = None) -> None:
        self.token_map = {'comment': Comment,
                          'section_name': Section,
                          'entry': Entry}
        if token_map is not None:
            self.token_map |= token_map

    def parse_action(self, tokens: pp.ParseResults) -> Token:
        name = tokens.get_name()
        token_class = self.token_map[name]
        return token_class(*tokens)


# region [Main_Exec]

if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union, Literal, TypeVar, Optional
from pathlib import Path
import logging
from datetime import datetime, timedelta
import os
# * Third Party Imports --------------------------------------------------------------------------------->
from frozendict import frozendict

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import MaxError, MinError, ValueValidationError, NotLoggingLevelError
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.general_helper.conversion import bytes2human, human2bytes, str_to_bool, seconds2human, human2timedelta
from gidapptools.general_helper.import_helper import is_importable
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals

try:
    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_config.conversion.conversion_table import ConversionTable

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
PYSIDE6_AVAILABLE = is_importable("PySide6")
YARL_AVAILABLE = is_importable("yarl")
THIS_FILE_DIR = Path(__file__).parent.absolute()

get_dummy_profile_decorator_in_globals()
# endregion [Constants]


class ConfigValueConverter(ABC):
    __slots__ = ("conversion_table",)
    is_standard_converter: bool = False
    value_typus: str = None
    value_typus_aliases: tuple[str] = tuple()
    named_arguments_options: frozendict[str, object] = frozendict()
    dependencies: list[str] = None

    def __init__(self, conversion_table: "ConversionTable") -> None:
        self.conversion_table = conversion_table

    def __init_subclass__(cls) -> None:
        if cls.value_typus in {None, ..., "", MiscEnum.NOTHING}:
            raise ValueError(f"{cls.__name__!r} has an invalid value_typus({cls.value_typus!r})")

    @property
    def encode(self):
        return self.to_config_value

    @property
    def decode(self):
        return self.to_python_value

    def validate_value(self, value, entry_name: str, section_name: str, config) -> None:
        ...

    @abstractmethod
    def to_config_value(self, value: Any) -> str:
        ...

    @abstractmethod
    def to_python_value(self, value: str) -> Any:
        ...

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}'


class IntegerConfigValueConverter(ConfigValueConverter):
    __slots__ = ("minimum", "maximum", "on_min_error", "on_max_error")
    is_standard_converter: bool = True
    value_typus = "integer"
    value_typus_aliases = ("int",)
    is_default: bool = True

    def __init__(self, conversion_table: "ConversionTable", minimum: int = None, maximum: int = None, on_min_error: Literal["raise", "set_minimum"] = "raise", on_max_error: Literal["raise", "set_maximum"] = "raise") -> None:
        super().__init__(conversion_table)
        self.minimum = minimum
        self.maximum = maximum
        self.on_min_error = on_min_error
        self.on_max_error = on_max_error

    def to_config_value(self, value: int) -> str:
        try:
            self.validate_value(value, None, None, None)
        except MinError:
            if self.on_min_error == "raise":
                raise
            if self.on_min_error == "set_minimum":
                value = self.minimum

        except MaxError:
            if self.on_max_error == "raise":
                raise
            if self.on_max_error == "set_maximum":
                value = self.maximum

        if value is None:
            return ""

        return str(value)

    def to_python_value(self, value: str) -> Union[int, None]:
        if value is None:
            return None

        converted_value = int(value)
        if self.minimum:
            converted_value = max(converted_value, self.minimum)
        if self.maximum:
            converted_value = min(converted_value, self.maximum)
        return converted_value

    def validate_value(self, value: Optional[int], entry_name: str, section_name: str, config) -> None:
        if value is None:
            return

        if self.minimum is not None and value < self.minimum:
            raise MinError(value, self.minimum, converter=self, section_name=section_name, entry_name=entry_name, config=config)

        if self.maximum is not None and value > self.maximum:
            raise MaxError(value, self.maximum, converter=self, section_name=section_name, entry_name=entry_name, config=config)


class FloatConfigValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "float"
    value_typus_aliases = ("floating_point", "double")

    def to_config_value(self, value: float) -> str:
        if value is None:
            return ""
        return str(value)

    def to_python_value(self, value: str) -> Union[float, None]:
        if value is None:
            return None
        return float(value)


class BooleanConfigValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "boolean"
    value_typus_aliases = ("bool",)
    true_string_value = "yes"
    false_string_value = "no"

    def to_config_value(self, value: bool) -> str:
        if value is None:
            return ""
        if value is True:
            return self.true_string_value

        if value is False:
            return self.false_string_value

    def to_python_value(self, value: str) -> Union[bool, None]:
        if value is None:
            return None
        return str_to_bool(str(value))


class StringConfigValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "string"
    value_typus_aliases = ("str",)

    def to_config_value(self, value: str) -> str:
        if value is None:
            return ""
        return value

    def to_python_value(self, value: str) -> Union[str, None]:
        if value is None:
            return None
        return value


class DateTimeConfigValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "datetime"

    def to_config_value(self, value: datetime) -> str:
        if value is None:
            return ""
        return value.isoformat(timespec="seconds")

    def to_python_value(self, value: str) -> Union[datetime, None]:
        if value is None:
            return None
        return datetime.fromisoformat(value)


class TimedeltaConfigValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "timedelta"

    def to_config_value(self, value: timedelta) -> str:
        if value is None:
            return ""
        return seconds2human(value)

    def to_python_value(self, value: str) -> Union[timedelta, None]:
        if value is None:
            return None
        return human2timedelta(value)


class PathConfigValueConverter(ConfigValueConverter):
    __slots__ = ("resolve",)
    is_standard_converter: bool = True
    value_typus = "path"
    value_typus_aliases = ("fs_path",)

    def __init__(self, conversion_table: "ConversionTable", resolve: bool = True) -> None:
        super().__init__(conversion_table)
        self.resolve = resolve

    def to_config_value(self, value: Path) -> str:
        if value is None:
            return ""
        path = Path(value)
        if self.resolve is True:
            path = path.resolve()
        return path.as_posix()

    def to_python_value(self, value: str) -> Union[Path, None]:
        if value is None:
            return None
        path = Path(value.strip())
        if self.resolve is True:
            path = Path(os.path.expandvars(path)).resolve()

        return path


class FileSizeConfigValueConverter(ConfigValueConverter):
    __slots__ = ()
    is_standard_converter: bool = True
    value_typus: str = "file_size"

    def to_config_value(self, value: int) -> str:
        if value is None:
            return ""
        return bytes2human(value)

    def to_python_value(self, value: str) -> Union[int, None]:
        if value is None:
            return None

        return human2bytes(value)


SUB_TYPUS_TYPE = TypeVar("SUB_TYPUS_TYPE")


class ListConfigValueConverter(ConfigValueConverter):
    __slots__ = ("sub_typus", "split_char")
    is_standard_converter: bool = True
    value_typus = "list"

    def __init__(self, conversion_table: "ConversionTable", sub_typus: str = "string", split_char: str = ",") -> None:
        super().__init__(conversion_table)
        self.sub_typus = sub_typus
        self.split_char = split_char

    def to_config_value(self, value: list) -> str:
        if value is None:
            return ""
        sub_converter: "ConfigValueConverter" = self.conversion_table.converters[self.sub_typus](self.conversion_table)
        return f"{self.split_char} ".join(sub_converter.to_config_value(item) for item in value)

    def to_python_value(self, value: str) -> Union[list[SUB_TYPUS_TYPE], None]:
        if value is None:
            return None
        sub_converter: "ConfigValueConverter" = self.conversion_table.converters[self.sub_typus](self.conversion_table)
        return [sub_converter.to_python_value(item) for item in value.split(self.split_char) if item.strip()]


class LoggingLevelValueConverter(ConfigValueConverter):
    __slots__ = tuple()
    is_standard_converter: bool = True
    value_typus = "logging_level"

    def validate_value(self, value, entry_name: str, section_name: str, config) -> None:
        if not isinstance(value, (str, int)):
            raise NotLoggingLevelError(value=value, converter=self, entry_name=entry_name, section_name=section_name, config=config)

        if isinstance(value, int) and value not in logging._levelToName:
            raise NotLoggingLevelError(value=value, converter=self, entry_name=entry_name, section_name=section_name, config=config)

        if isinstance(value, str) and value not in logging._nameToLevel:
            raise NotLoggingLevelError(value=value, converter=self, entry_name=entry_name, section_name=section_name, config=config)

    def to_config_value(self, value: Any, **named_arguments) -> str:
        if value is None:
            return ""

        if isinstance(value, int):

            value = logging._levelToName[value]

        if isinstance(value, str):
            value = value.upper()

        self.validate_value(value, None, None, None)

        return value

    def to_python_value(self, value: str, **named_arguments) -> Union[int, None]:

        value = value.upper()

        self.validate_value(value, None, None, None)

        return logging._nameToLevel[value]


if PYSIDE6_AVAILABLE is True:
    from PySide6.QtGui import QColor

    class QColorValueConverter(ConfigValueConverter):
        __slots__ = tuple()
        is_standard_converter: bool = True
        value_typus = "qcolor"
        dependencies: list[str] = ["PySide6"]

        def to_config_value(self, value: Any, **named_arguments) -> str:
            if value is None:
                return ""
            return ', '.join(str(i) for i in value.getRgb())

        def to_python_value(self, value: str, **named_arguments) -> Union[QColor, None]:
            if value is None:
                return None
            r, g, b, a = (int(i.strip()) for i in value.split(',') if i.strip())
            return QColor(r, g, b, a)


if YARL_AVAILABLE is True:
    from yarl import URL

    class URLValueConverter(ConfigValueConverter):
        __slots__ = tuple()
        is_standard_converter: bool = True
        value_typus = "url"
        dependencies: list[str] = ["yarl"]

        def to_config_value(self, value: Union[URL, None]) -> str:
            if value is None:
                return ""
            return value.human_repr()

        def to_python_value(self, value: Union[str, None]) -> Union[URL, None]:
            if value is None:
                return None
            return URL(value)


def get_standard_converter() -> tuple[type["ConfigValueConverter"]]:
    return tuple(sc for sc in ConfigValueConverter.__subclasses__() if sc.is_standard_converter is True)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Union, Iterable, Optional
from pathlib import Path
from colorsys import hls_to_rgb, hsv_to_rgb, rgb_to_hls, rgb_to_hsv
from functools import cached_property

# * Third Party Imports --------------------------------------------------------------------------------->
import attr

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gidcolor.preset_colors.web_colors import _load_webcolors_data
import numpy as np
try:
    from PySide6.QtGui import QColor
    PYSIDE6_AVAILABLE = True
except ImportError:
    PYSIDE6_AVAILABLE = False

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


from warnings import warn_explicit, warn
from . import color

warn(message=f"Submodule {__name__!r} is deprecated, please use submodule {color.__name__!r}.",
     category=DeprecationWarning,
     stacklevel=2)


# endregion [Constants]


INT_OR_FLOAT = Union[int, float]

COLOR_FLOAT_TYPE = Union[tuple[float, float, float], tuple[float, float, float, float]]

COLOR_INT_TYPE = Union[tuple[int, int, int], tuple[int, int, int, float]]


def _validate_color_float(instance: object, attribute: attr.Attribute, value: Optional[float]) -> None:

    if 0.0 > value > 1.0:

        raise ValueError(f"{attribute.name!r}-value can only be between 0 and 1, not {value!r}.")


def _alpha_converter(alpha: Union[int, float] = None) -> float:
    if alpha is None:
        return 1.0

    return float(alpha)


def rgb_to_hex(r, g, b):
    parts = []
    for part in [r, g, b]:
        if isinstance(part, float):
            part = int(255 * part)
        parts.append(part)
    return '#' + ''.join(f"{p:X}"for p in parts)


def rgb_float_to_rgb_int(in_rgb_float: tuple[float, float, float]) -> tuple[int, int, int]:
    return tuple(int(255 * i) for i in in_rgb_float)


class ColorTypus(Enum):
    RGB = auto()
    HLS = auto()
    HSV = auto()


class BaseColor(ABC):
    float_round_n: int = 4

    def __iter__(self):
        return (getattr(self, field.name) for field in attr.fields(self.__class__) if field.metadata.get("is_core_color", False) is True)

    @abstractmethod
    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    @abstractmethod
    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        ...

    @abstractmethod
    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    @abstractmethod
    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        ...

    def as_hex(self, include_alpha: bool = True) -> str:
        _rgb = self.as_rgb_int(include_alpha=include_alpha, alpha_as_int=True)
        _hex = rgb_to_hex(*_rgb[:3])
        if include_alpha is True and len(_rgb) == 4:
            _hex += f"{_rgb[-1]:X}"

        return _hex

    @cached_property
    def qcolor(self) -> Optional["QColor"]:
        if PYSIDE6_AVAILABLE:
            return QColor(*self.as_rgb_int(True, True))

    def to_q_style_rgba_string(self,) -> str:
        values = (str(i) for i in self.as_rgb_int(alpha_as_int=True))
        return "rgba(" + ', '.join(values) + ')'

    def __repr__(self) -> str:
        class_name = self.__class__.__name__.removesuffix("Color")
        _out = f"{class_name}(" + ', '.join(repr(x) for x in self)
        if self.name is not None:
            _out += f', name={self.name!r}'

        if self.aliases:
            _out += f", aliases={tuple(self.aliases)!r}"
        return _out + ')'

    def with_alpha(self, alpha: Union[int, float]) -> "BaseColor":
        if isinstance(alpha, int) or alpha > 1:
            alpha = round(alpha / 255, self.float_round_n)
        return self.__class__(*list(self)[:-1], alpha)


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class RGBColor(BaseColor):
    red: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    green: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    blue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=_validate_color_float, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))

    np_value: "np.ndarray" = attr.ib()

    @np_value.default
    def _np_value_default(self):
        return np.asfarray(self.as_rgb_float(True, False), dtype=np.float32)

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in list(self)[:-1]]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hls(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hsv(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class HLSColor(BaseColor):
    hue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    saturation: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    lightness: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=_validate_color_float, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))

    np_value: np.ndarray = attr.ib()

    @np_value.default
    def _np_value_default(self):
        return np.asfarray(self.as_rgb_float(True, False), dtype=np.float32)

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in hls_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in hls_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hsv(*hls_to_rgb(*list(self)[:-1]))]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out


@attr.s(auto_attribs=True, auto_detect=True, slots=True, weakref_slot=True, frozen=True, order=True, repr=False)
class HSVColor(BaseColor):
    hue: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    saturation: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})
    value: float = attr.ib(validator=_validate_color_float, metadata={'is_core_color': True})

    alpha: float = attr.ib(default=1.0, converter=_alpha_converter, validator=attr, metadata={'is_core_color': True})

    name: str = attr.ib(default=None)
    aliases: Iterable[str] = attr.ib(factory=set, converter=attr.converters.default_if_none(factory=set))
    np_value: np.ndarray = attr.ib()

    @np_value.default
    def _np_value_default(self):
        return np.asfarray(self.as_rgb_float(True, False), dtype=np.float32)

    def as_hsv(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = list(self)[:-1]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_hls(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in rgb_to_hls(*hsv_to_rgb(*list(self)[:-1]))]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return _out

    def as_rgb_float(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_FLOAT_TYPE:
        _out = [round(col, self.float_round_n) for col in hsv_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)

    def as_rgb_int(self, include_alpha: bool = True, alpha_as_int: bool = False) -> COLOR_INT_TYPE:
        _out = [int(sub_col * 255) for sub_col in hsv_to_rgb(*list(self)[:-1])]
        if include_alpha is True:
            if alpha_as_int is True:
                _out.append(int(self.alpha * 255))
            else:
                _out.append(self.alpha)
        return tuple(_out)


color_class_map: dict[str, type[BaseColor]] = {"rgb": RGBColor,
                                               "hls": HLSColor,
                                               "hsv": HSVColor}


class ColorRegistry:
    color_typus = ColorTypus

    def __init__(self) -> None:
        self.colors_by_name: dict[str:RGBColor] = {}

    def _add_color(self, color: RGBColor) -> None:
        if color.name:
            self.colors_by_name[color.name.casefold()] = color
        for alias in color.aliases:
            self.colors_by_name[alias.casefold()] = color

    def color_factory(self, value: tuple, typus: ColorTypus, name: str = None, aliases: Iterable[str] = None):
        if len(value) == 3:
            value = (value[0], value[1], value[2], 1.0)

        if typus is ColorTypus.RGB:
            _values = []
            for v in value:
                if isinstance(v, int) and v != 0:
                    v = round(v / 255, BaseColor.float_round_n)
                _values.append(v)

            color = RGBColor(*_values, name=name, aliases=aliases)

        elif typus is ColorTypus.HLS:
            ...

        elif typus is ColorTypus.HSV:
            ...

        self._add_color(color)
        return color

    def __call__(self, value: tuple, typus: ColorTypus, name: str = None, aliases: Iterable[str] = None) -> Any:
        return self.color_factory(value=value, typus=typus, name=name, aliases=aliases)

    def get_color_by_name(self, name: str) -> "RGBColor":
        if len(self.colors_by_name) == 0:
            for _p_color in _load_webcolors_data():
                Color(**_p_color, typus=Color.color_typus.RGB)

        return self.colors_by_name[name.casefold()]


xyx = []
Color = ColorRegistry()
for p_color in _load_webcolors_data():
    xyx.append(Color.color_factory(**p_color, typus=Color.color_typus.RGB))

# region [Main_Exec]
if __name__ == '__main__':
    from pympler.asizeof import asizeof
    from gidapptools.general_helper.conversion import bytes2human
    import inspect
    print(f"{bytes2human(asizeof(xyx[0]))=}")
    print(f"{bytes2human(asizeof(xyx[0].np_value))=}")
    print(f"{xyx[0].np_value=}")

    print(f"{RGBColor.f_code}")


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from enum import Enum, StrEnum
from pathlib import Path
from typing import TYPE_CHECKING, Union, Generator
from datetime import datetime, timezone, timedelta, tzinfo
from functools import total_ordering
import pytz
import traceback
from collections import defaultdict
import copy
from threading import Lock

from math import ceil
# * Third Party Imports --------------------------------------------------------------------------------->


# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import DateTimeFrameTimezoneError, NotUtcDatetimeError
from gidapptools.gid_logger.logger import get_logger

import sys
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
log = get_logger(__name__)
# endregion [Constants]


class DatetimeFmt(StrEnum):
    STANDARD = "%Y-%m-%d %H:%M:%S"
    FILE = "%Y-%m-%d_%H-%M-%S"
    LOCAL = "%x %X"

    STANDARD_TZ = "%Y-%m-%d %H:%M:%S %Z"
    FILE_TZ = "%Y-%m-%d_%H-%M-%S_%Z"
    LOCAL_TZ = "%x %X %Z"


def get_all_timezone_names() -> tuple[str]:
    return tuple(sorted(pytz.all_timezones))


def get_all_timezones() -> tuple[pytz.BaseTzInfo]:
    return tuple(pytz.timezone(i) for i in get_all_timezone_names())


def get_all_timezone_names_by_region() -> dict[str, list[str]]:
    single_europe_names = {"Poland", "Portugal", "Iceland", "Turkey", "Eire", "GB", "GB-Eire", "EET", "CET", "WET", "Greenwich"}

    single_asia_names = {"Hongkong", "Japan", "Singapore", "ROK", "ROC", "PRC", "Turkey", "Iran", "Israel"}

    single_africa_names = {"Libya", "Egypt"}

    single_america_names = {"Cuba", "Jamaica", "Navajo", "HST", "MST"}

    single_pacific_names = {"Kwajalein", "NZ", "NZ-CHAT", "HST"}

    all_single_names = single_europe_names.union(single_asia_names, single_africa_names, single_america_names, single_pacific_names)

    region_map = defaultdict(list)
    for time_zone_name in get_all_timezone_names():
        if time_zone_name.casefold().startswith("etc/") or "/" not in time_zone_name:
            if time_zone_name in all_single_names:
                if time_zone_name in single_europe_names:
                    region_map["Europe"].append(time_zone_name)

                if time_zone_name in single_asia_names:
                    region_map["Asia"].append(time_zone_name)

                if time_zone_name in single_africa_names:
                    region_map["Africa"].append(time_zone_name)

                if time_zone_name in single_america_names:
                    region_map["America"].append(time_zone_name)

                if time_zone_name in single_pacific_names:
                    region_map["Pacific"].append(time_zone_name)
            else:
                region_map["misc"].append(time_zone_name)

        else:
            for region_name in time_zone_name.split("/")[:-1]:
                region_map[region_name].append(time_zone_name)

    return region_map


def get_all_timezones_by_region() -> dict[str, list[pytz.BaseTzInfo]]:
    new_map = {}
    time_zone_map = get_all_timezone_names_by_region()
    for region in time_zone_map:
        new_map[region] = [pytz.timezone(i) for i in time_zone_map[region]]

    return new_map


def get_all_timezones_by_offset_hours(include_dst_timezone: bool = True) -> dict[int, tuple[pytz.BaseTzInfo]]:
    now = datetime.now()
    now_three_years_ago = now.replace(year=now.year - 3)
    _out = defaultdict(list)

    for time_zone in get_all_timezones():
        if include_dst_timezone is False and isinstance(time_zone, pytz.tzinfo.DstTzInfo) and time_zone._utc_transition_times[-1] >= now_three_years_ago:

            continue
        try:
            offset = time_zone.utcoffset(now)
            offset_hours = int(offset.total_seconds() // 3600)
            _out[offset_hours].append(time_zone)
        except Exception as e:
            log.error(e, exc_info=True)

    return {k: tuple(v) for k, v in _out.items()}


def get_aware_now(tz: timezone) -> datetime:
    return datetime.now(tz=tz)


def get_utc_now() -> datetime:
    return get_aware_now(tz=timezone.utc)


@ total_ordering
class DateTimeFrame:
    __slots__ = ("_start",
                 "_end")

    def __init__(self, start: datetime, end: datetime) -> None:
        self._start = start
        self._end = end
        self._validate()

    def _validate(self) -> None:
        if self.start.tzinfo is None or self.end.tzinfo is None:
            raise DateTimeFrameTimezoneError(self, self.start.tzinfo, self.end.tzinfo, 'start time and end time need to be timezone aware')
        if self.start.tzinfo != self.end.tzinfo:
            raise DateTimeFrameTimezoneError(self, self.start.tzinfo, self.end.tzinfo, 'start time and end time do not have the same timezone')

        if self.start > self.end:
            raise ValueError(f"start {self.start!r} cannot be after end ({self.end!r}) for {self.__class__.__name__!r}.")

    @property
    def start(self) -> datetime:
        return self._start

    @property
    def end(self) -> datetime:
        return self._end

    @ property
    def delta(self) -> timedelta:
        return self.end - self.start

    @ property
    def tzinfo(self) -> timezone:
        return self.start.tzinfo

    def with_start(self, new_start: datetime) -> Self:
        return self.__class__(new_start, self.end)

    def with_end(self, new_end: datetime) -> Self:
        return self.__class__(self.start, new_end)

    def modify_start(self, delta: timedelta) -> Self:
        return self.__class__(self.start + delta, self.end)

    def modify_end(self, delta: timedelta) -> Self:
        return self.__class__(self.start, self.end + delta)

    def modify_start_and_end(self, delta: timedelta) -> Self:
        return self.__class__(self.start + delta, self.end + delta)

    def iter_equal_steps(self, amount_of_steps: int) -> Generator[datetime, None, None]:
        step_size = self.delta / (amount_of_steps - 1)
        print(f"{step_size=}")

        for i in range(amount_of_steps - 1):
            yield self.start + (step_size * i)
        yield self.end

    def __sub__(self, other: object) -> Self:
        if isinstance(other, timedelta):
            return self.modify_end(-other)

        if isinstance(other, datetime) and other in self:
            if other <= self.end:
                return self.with_end(other)

            if other >= self.start:
                return self.with_start(other)

        return NotImplemented

    def __add__(self, other: object) -> Self:
        if isinstance(other, timedelta):
            return self.modify_end(other)

        if isinstance(other, datetime):
            if other > self.end:
                return self.with_end(other)

            if other < self.start:
                return self.with_start(other)

            if other in self:
                return copy.copy(self)

        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DateTimeFrame):
            return self.start == other.start and self.end == other.end

        if isinstance(other, timedelta):
            return self.delta == other

        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, datetime):
            return self.start < other

        if isinstance(other, DateTimeFrame):
            return self.start < other.start

        if isinstance(other, timedelta):
            return self.delta < other

        return NotImplemented

    def __contains__(self, other: object) -> bool:
        if isinstance(other, datetime):
            return self.start <= other <= self.end

        if isinstance(other, DateTimeFrame):
            return self.start <= other.start and self.end >= other.end

        return NotImplemented

    def __copy__(self) -> Self:
        new_instance = self.__class__.__new__(self.__class__)

        for attr_name in self.__slots__:
            if attr_name not in {"weakref", }:
                setattr(new_instance, attr_name, getattr(self, attr_name))

        return new_instance

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(start={self.start!r}, end={self.end!r})"

    def __str__(self) -> str:
        return f"{self.start.isoformat(sep=' ')} until {self.end.isoformat(sep=' ')}"


ZERO_TIMEDELTA = timedelta()


def calculate_utc_offset(utc_datetime: datetime, local_datetime: datetime, offset_class: type[timezone] = timezone, as_pytz_timezone: bool = False) -> Union[timezone, pytz.BaseTzInfo]:
    if utc_datetime.tzinfo.tzname(None) != "UTC" and utc_datetime.tzinfo != timezone.utc and utc_datetime.tzinfo.utcoffset(None) != ZERO_TIMEDELTA:
        raise NotUtcDatetimeError(utc_datetime)
    difference_seconds = (local_datetime.replace(tzinfo=timezone.utc) - utc_datetime).total_seconds()

    offset_timedelta = timedelta(seconds=int(difference_seconds))
    offset_hours = offset_timedelta.total_seconds() / (60 * 60)

    offset_hours = round(offset_hours)

    prefix = "" if (offset_hours * (-1)) < 0 else "+"

    name = "Etc/GMT" + prefix + str(offset_hours * (-1))

    if as_pytz_timezone is True:
        return pytz.timezone(name)
    return offset_class(offset=timedelta(hours=offset_hours), name=name)


def pytz_timezone_from_utc_offset(in_offset_hours: int, only_named_timezones: bool = False, allow_dst_timezone: bool = True) -> tuple[pytz.BaseTzInfo]:
    ...


# region [Main_Exec]
if __name__ == '__main__':
    x = datetime.now(tz=timezone.utc) - timedelta(hours=8)
    y = datetime.now(tz=pytz.timezone("America/Los_Angeles")) - timedelta(hours=8)

    print(x.isoformat())
    print(y.isoformat())


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform

import subprocess
import inspect

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final, Callable,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED

from colorsys import hsv_to_rgb, rgb_to_hls, rgb_to_hsv, hls_to_rgb

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def float_to_degrees(in_value: float) -> float:
    return (in_value * 360) % 360


def rgb_to_hsl(r: float, g: float, b: float) -> tuple[float, float, float]:
    h, l, s = rgb_to_hls(r, g, b)
    return h, s, l


def hsl_to_rgb(h: float, s: float, l: float) -> tuple[float, float, float]:
    return hls_to_rgb(h, l, s)


def rgb_float_to_rgb_int(in_rgb_float: tuple[float]) -> tuple[int]:
    return tuple(round(255 * i) for i in in_rgb_float)


def rgb_int_to_rgb_float(in_rgb_int: tuple[float]) -> tuple[int]:
    return tuple(float(i / 255) for i in in_rgb_int)


def rgb_float_to_hex(in_rgb_float: tuple[float]) -> str:
    values = rgb_float_to_rgb_int(in_rgb_float)
    return '#' + ''.join(f"{p:02X}" for p in values)


def hex_to_rgb_float(in_hex: str) -> tuple[float]:
    cleaned_hex = in_hex.removeprefix("#")
    rgb_int_values = tuple(bytes.fromhex(cleaned_hex))
    return rgb_int_to_rgb_float(rgb_int_values)


# region [Main_Exec]
if __name__ == '__main__':
    import numpy as np
    x = (0.0, 0.6901960784313725, 1.0)

    y = rgb_to_hsv(*x)

    z = ((y[0] + 0.5) - 1.0, (y[1] / 4) * 3, (y[2] / 4) * 3)

    print(f"{z=}")
    zz = hsv_to_rgb(*z)
    print(f"{rgb_float_to_hex(zz)=}")

    # x = "#00B0FF"
    # print(f"{hex_to_rgb_float(x)=}")
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from enum import Enum, auto
from types import FunctionType
from pathlib import Path
from warnings import warn_explicit
from functools import wraps

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import MiscEnum

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class DeprecationWarningTypus(Enum):
    ARGUMENT = auto()


def _replace_default_argument_w_NOTHING(func, arg_name: str, *args, **kwargs) -> inspect.BoundArguments:
    func_sig = inspect.signature(func)
    new_arg = func_sig.parameters[arg_name].replace(default=MiscEnum.NOTHING)
    new_params = [param_value if param_name != arg_name else new_arg for param_name, param_value in func_sig.parameters.items()]
    new_sig = func_sig.replace(parameters=new_params)
    bound_args = new_sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args


def _make_deprication_warning(typus: DeprecationWarning, **kwargs) -> None:
    func: FunctionType = kwargs["func"]
    if typus is DeprecationWarningTypus.ARGUMENT:

        needed = {"arg_name", "func_name", "func"}
        if any(kwarg_name not in kwargs for kwarg_name in needed):
            missing = [pos_missing for pos_missing in needed if kwargs.get(pos_missing, None) is None]
            raise AttributeError(f"missing needed kwargs {', '.join(repr(i) for i in missing)}.")

        message_parts = [f"The argument {kwargs.get('arg_name')!r} for '{kwargs.get('func_name')}()' is deprecated"]
        if kwargs.get('not_used', False):
            message_parts.append("it is not used anymore in the actual function (does not do anything)")
        if kwargs.get('alternative_arg_name', None):
            message_parts.append(f"use the alternative {kwargs.get('alternative_arg_name')!r}")
        message = ', '.join(message_parts) + '.'

    warn_explicit(message=message, category=DeprecationWarning, filename=func.__code__.co_filename, lineno=func.__code__.co_firstlineno, module=func.__module__)
    # warn(message=message, category=DeprecationWarning, stacklevel=3)


def deprecated_argument(arg_name: str, alternative_arg_name: str = None, not_used: bool = True):

    def _wrapper(func):
        func_name = func.__qualname__ or func.__name__

        @wraps(func)
        def _wrapped(*args, **kwargs):
            new_bound_args = _replace_default_argument_w_NOTHING(func, arg_name, *args, **kwargs)
            args = new_bound_args.args
            kwargs = new_bound_args.kwargs
            if new_bound_args.arguments[arg_name] is not MiscEnum.NOTHING:
                _make_deprication_warning(DeprecationWarningTypus.ARGUMENT, func=func, arg_name=arg_name, func_name=func_name, not_used=not_used, alternative_arg_name=alternative_arg_name)
            return func(*args, **kwargs)

        return _wrapped

    return _wrapper


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
import sys
import inspect
from typing import Any, Callable
from pathlib import Path
from functools import partial

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.helper import meta_data_from_path
from gidapptools.meta_data.config_kwargs import ConfigKwargs
from gidapptools.general_helper.conversion import str_to_bool
from gidapptools.meta_data.meta_info.meta_info_item import MetaInfo
from gidapptools.abstract_classes.abstract_meta_factory import AbstractMetaFactory

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class License:
    special_licenses: dict[str, "License"] = {}

    def __init__(self, name: str, osi_approved: bool, description: str = None):
        self.name = name
        self.osi_approved = osi_approved
        self.description = description or ""

    @classmethod
    def from_classifier_string(cls, classifier_string: str):
        parts = [i.strip() for i in classifier_string.split("::")]
        osi_approved = True if "OSI Approved" in parts else False
        name = parts[-1]
        try:
            return cls.special_licenses[name]
        except KeyError:
            return cls(name=name, osi_approved=osi_approved)

    @classmethod
    def add_to_special_licenses(cls, special_license: "License"):
        cls.special_licenses[special_license.name] = special_license

    def __str__(self) -> str:
        return self.name


License.add_to_special_licenses(License(name="MIT License", osi_approved=True, description=""))
License.add_to_special_licenses(License(name="Unknown", osi_approved=False, description="License is missing or unknown."))


class MetaInfoFactory(AbstractMetaFactory):
    product_class = MetaInfo
    prefix_arg_getters = '_arg_get_'
    is_dev_env_name = 'IS_DEV'
    is_code_runner_env_name = 'IS_CODE_RUNNER'

    def __init__(self, config_kwargs: ConfigKwargs) -> None:
        super().__init__(config_kwargs=config_kwargs)

        self.init_path = Path(config_kwargs.get('init_path'))
        self.package_metadata = meta_data_from_path(self.init_path)
        self.package_metadata['app_name'] = self.package_metadata.pop('name')
        self.package_metadata['app_author'] = self.package_metadata.pop('author')

        self.arg_getters_map: dict[str, Callable] = None
        self.needed_arg_names: list[str] = None

    def setup(self) -> None:
        self.arg_getters_map = self._collect_arg_getters_map()
        self.needed_arg_names = self._retrieve_needed_arg_names()
        self.is_setup = True

    def _collect_arg_getters_map(self) -> dict[str, Callable]:
        arg_getters_map = {}
        for meth_name, meth_obj in inspect.getmembers(self, inspect.ismethod):
            if meth_name.startswith(self.prefix_arg_getters):
                arg_name = meth_name.removeprefix(self.prefix_arg_getters)
                arg_getters_map[arg_name] = meth_obj
        return arg_getters_map

    def _retrieve_needed_arg_names(self) -> list[str]:
        parameters = inspect.signature(self.product_class).parameters
        return list(parameters)

    def _build_meta_info_args(self) -> dict[str, Any]:
        meta_info_kwargs = {arg_name: self.config_kwargs.get(arg_name, None) for arg_name in self._retrieve_needed_arg_names()}
        for arg_name, arg_value in meta_info_kwargs.items():

            if arg_value is not None:
                continue

            arg_value_getter = self.arg_getters_map.get(arg_name, partial(self._default_arg_getter, arg_name))

            meta_info_kwargs[arg_name] = arg_value_getter()
        return {k: v for k, v in meta_info_kwargs.items() if v is not None}

    def _default_arg_getter(self, arg_name: str) -> Any:

        return self.package_metadata.get(arg_name.casefold())

    def _arg_get_app_license(self) -> License:
        license_classifier = next((i for i in self.package_metadata.get("classifier", []) if i.startswith("License")), "License :: Unknown")
        return License.from_classifier_string(license_classifier)

    def _arg_get_other_urls(self) -> dict[str, str]:
        _out = {}
        for item in self.package_metadata.get('project-url', []):
            if item.startswith("Source"):
                continue

            parts = [i.strip() for i in item.split(",") if i.strip()]
            if len(parts) != 2:
                continue
            if not parts[1].startswith("http"):
                continue
            _out[parts[0]] = parts[1]
        return _out

    def _arg_get_url(self) -> str:
        url_list = self.package_metadata.get('project-url', [])

        url_text = next((i for i in url_list if i.startswith("Source")), "")

        parts = map(lambda x: x.strip(), url_text.split(','))
        for part in parts:
            if part.startswith('http'):
                return part

    def _arg_get_is_dev(self) -> bool:
        is_dev_string = os.getenv(self.is_dev_env_name, '0').casefold()
        return str_to_bool(is_dev_string, strict=True) or sys.flags.dev_mode

    def _arg_get_is_gui(self) -> bool:
        if str_to_bool(os.getenv(self.is_code_runner_env_name, '0'), strict=True) is True:
            return False
        stdout_is_gui = sys.stdout is None or sys.stdout.isatty() is False
        stderr_is_gui = sys.stderr is None or sys.stderr.isatty() is False
        stdin_is_gui = sys.stdin is None or sys.stdin.isatty() is False
        return all([stdout_is_gui, stderr_is_gui, stdin_is_gui])

    def _build(self) -> MetaInfo:
        if self.is_setup is False:
            # TODO: maybe raise error instead
            self.setup()
        meta_info_kwargs = self._build_meta_info_args()
        instance = self.product_class(**meta_info_kwargs)
        return instance


# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
import inspect
from string import punctuation, ascii_lowercase
from typing import Any, Union, Literal, Mapping, Callable, Iterable, Optional
from pathlib import Path
from textwrap import dedent

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as ppa

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.enums import StringCase
from gidapptools.gid_warning.experimental import mark_experimental

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]

# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()


# endregion [Constants]

STRING_CASE_FUNC_TYPE = Callable[[Iterable[str]], str]


# TODO: Rewrite as normal class/Module-Singleton
class _StringCaseConverter:
    SNAKE = StringCase.SNAKE
    SCREAMING_SNAKE = StringCase.SCREAMING_SNAKE
    CAMEL = StringCase.CAMEL
    PASCAL = StringCase.PASCAL
    KEBAP = StringCase.KEBAP
    SPLIT = StringCase.SPLIT
    CLASS = StringCase.CLASS
    TITLE = StringCase.TITLE
    BLOCK_UPPER = StringCase.BLOCK_UPPER

    split_pascal_case_regex = re.compile(r"(?<!\_)(\B[A-Z])")
    snake_case_to_pascal_case_regex = re.compile(r"(_|^)(\w)")
    _word_list_split_chars = frozenset({'-', '_', ' '})

    __slots__ = ("_split_grammar",
                 "_word_list_split_regex",
                 "_dispatch_table",
                 "_bad_chars")

    def __init__(self) -> None:
        self._split_grammar: Optional[ppa.ParserElement] = None

        self._word_list_split_regex = None

        self._dispatch_table: Optional[dict[str, STRING_CASE_FUNC_TYPE]] = None
        self._bad_chars: Optional[frozenset[str]] = None

    @property
    def bad_chars(self) -> set[str]:
        if self._bad_chars is None:
            self._bad_chars = frozenset({c for c in punctuation if c not in self._word_list_split_chars})
        return self._bad_chars

    @property
    def word_list_split_regex(self) -> re.Pattern:
        if self._word_list_split_regex is None:
            self._word_list_split_regex = re.compile(r'|'.join(list(self._word_list_split_chars) + [r"(?=[A-Z])"]))
        return self._word_list_split_regex

    @property
    def dispatch_table(self) -> dict[StringCase, STRING_CASE_FUNC_TYPE]:
        if self._dispatch_table is None:
            self._dispatch_table = {}
            for meth_name, meth_obj in inspect.getmembers(self):
                if meth_name.startswith('_to_') and meth_name.endswith('_case'):
                    key_name = meth_name.removeprefix('_to_').removesuffix('_case')
                    self._dispatch_table[StringCase(key_name)] = meth_obj
        return self._dispatch_table

    @property
    def split_grammar(self) -> ppa.ParserElement:
        if self._split_grammar is None:
            underscore = ppa.Literal('_').suppress()
            dash = ppa.Literal("-").suppress()
            all_upper_word = ppa.Regex(r"[A-Z]+(?![a-z])")
            all_lower_word = ppa.Word(ascii_lowercase, ascii_lowercase)
            title_word = ppa.Regex(r"[A-Z][a-z]+")
            number = ppa.Word(ppa.nums)
            words = (title_word | all_upper_word | all_lower_word | number).set_parse_action(lambda x: x[0].casefold())
            grammar = words | underscore | dash
            self._split_grammar = ppa.OneOrMore(grammar)
        return self._split_grammar

    def _to_word_list(self, in_string: str) -> list[str]:
        """
        :param in_string: str:
        """
        parts: ppa.ParseResults = self.split_grammar.parse_string(in_string, parse_all=True)

        return [word for word in parts if word]

    @staticmethod
    def _to_block_upper_case(word_list: Iterable[str]) -> str:
        """
        :param word_list: Iterable[str]:
        """
        return ''.join(word.upper() for word in word_list)

    @staticmethod
    def _to_snake_case(word_list: Iterable[str]) -> str:
        """
        :param word_list: Iterable[str]:
        """
        return '_'.join(word_list).casefold()

    @staticmethod
    def _to_camel_case(word_list: Iterable[str]) -> str:
        """
        :param word_list: Iterable[str]:
        """
        return word_list[0].casefold() + ''.join(item.title() for item in word_list[1:])

    @staticmethod
    def _to_pascal_case(word_list: Iterable[str]) -> str:
        """
        :param word_list: Iterable[str]:
        """
        return ''.join(item.title() for item in word_list)

    @staticmethod
    def _to_kebap_case(word_list: Iterable[str]) -> str:
        """

        :param word_list: Iterable[str]:

        """
        return '-'.join(word_list)

    @staticmethod
    def _to_screaming_snake_case(word_list: Iterable[str]) -> str:
        """

        :param word_list: Iterable[str]:

        """
        return '_'.join(word_list).upper()

    @staticmethod
    def _to_split_case(word_list: Iterable[str]) -> str:
        """

        :param word_list: Iterable[str]:

        """
        return ' '.join(word_list)

    @staticmethod
    def _to_title_case(word_list: Iterable[str]) -> str:
        """

        :param word_list: Iterable[str]:

        """
        return ' '.join(word.title() for word in word_list)

    @staticmethod
    def _to_upper_case(word_list: Iterable[str]) -> str:
        """

        :param word_list: Iterable[str]:

        """
        return ' '.join(word.upper() for word in word_list)

    def remove_bad_chars(self, in_string: str) -> str:
        new_string = str(in_string)
        for char in self.bad_chars:
            new_string: str = new_string.replace(char, "")

        return new_string

    def convert_to(self, in_string: str, target_case: Union[str, StringCase], clean_in_string: bool = False) -> str:
        """

        :param in_string: str:
        :param target_case: Union[str, StringCase]:

        """
        if clean_in_string is True:
            in_string = self.remove_bad_chars(in_string)
        target_case = StringCase(target_case) if isinstance(target_case, str) else target_case
        word_list = self._to_word_list(in_string)
        return self.dispatch_table.get(target_case)(word_list)


StringCaseConverter = _StringCaseConverter()
# _ = StringCaseConverter.dispatch_table
# _ = StringCaseConverter.split_grammar


def replace_by_dict(in_string: str, in_dict: dict[str, str]) -> str:
    """

    :param in_string: str:
    :param in_dict: dict[str, str]:

    """
    mod_string = in_string
    for key, value in in_dict.items():
        mod_string = mod_string.replace(key, value)
    return mod_string


def extract_by_map(in_string: str, extract_data: Union[Iterable[str], Mapping[str, str]]) -> Iterable[str]:
    """

    :param in_string: str:
    :param extract_data: Union[Iterable[str], Mapping[str, str]]:

    """
    parts = []
    re_pattern = re.compile(r'|'.join(extract_data))
    for match in re_pattern.finditer(in_string):
        matched_str = match.group()
        parts.append(matched_str)
    if isinstance(extract_data, Mapping):
        return [extract_data.get(part) for part in parts]
    return parts


SPACE_CLEANING_REGEX = re.compile(r" +")
NEWLINE_CLEANING_REGEX = re.compile(r"\n+")


def clean_whitespace(in_text: str, replace_newline: bool = False) -> str:
    """

    :param in_text: str:
    :param replace_newline: bool:  (Default value = False)

    """
    cleaned_text = SPACE_CLEANING_REGEX.sub(' ', in_text)
    if replace_newline is True:
        cleaned_text = NEWLINE_CLEANING_REGEX.sub(' ', cleaned_text)
    return cleaned_text


def shorten_string(in_text: str,
                   max_length: int,
                   shorten_side: Literal["right", "left"] = "right",
                   placeholder: str = '...',
                   clean_before: bool = True,
                   ensure_space_around_placeholder: bool = False,
                   split_on: str = r'\s|\n') -> str:
    """

    :param in_text: str:
    :param max_length: int:
    :param shorten_side: Literal["right", "left"]:  (Default value = "right")
    :param placeholder: str:  (Default value = "...")
    :param clean_before: bool:  (Default value = True)
    :param ensure_space_around_placeholder: bool:  (Default value = False)
    :param split_on: str:

    """
    max_length = int(max_length)
    if shorten_side.casefold() not in {"left", "right"}:
        raise ValueError(shorten_side)

    if clean_before is True:
        in_text = clean_whitespace(in_text, replace_newline=False)

    if len(in_text) <= max_length:
        return in_text

    if ensure_space_around_placeholder is True:
        placeholder = f" {placeholder.strip()}" if shorten_side == "right" else f"{placeholder.strip()} "

    max_length = max_length - len(placeholder)

    new_text = in_text[:max_length] if shorten_side == 'right' else in_text[-max_length:]
    if split_on == "any":
        split_on = r"."
    find_regex = re.compile(split_on)
    last_space_position = list(find_regex.finditer(new_text))

    return new_text[:last_space_position[-1].span()[0]].strip() + placeholder if shorten_side == 'right' else placeholder + new_text[last_space_position[0].span()[0]:].strip()


def split_quotes_aware(text: str,
                       split_chars: Iterable[str] = None,
                       quote_chars: Iterable[str] = None,
                       strip_parts: bool = True) -> list[str]:
    """Splits a string on but not if the separator char is inside of quotes.

    :param text: The string to split.
    :type text: str
    :param split_chars: The characters to split on. Defaults to `,`.
    :type split_chars: Iterable[str]
    :param quote_chars: The quote chars that should be considered real quotes. Defaults to `"` and `'`.
    :type quote_chars: Iterable[str]
    :param strip_parts: If each found substrin should be striped of preceding and trailing whitespace in the result. Defaults to True.
    :type strip_parts: bool
    :param text: str:
    :returns: The found sub-parts.
    :rtype: list[str]

    """
    split_chars = {','} if split_chars is None else set(split_chars)
    quote_chars = {"'", '"'} if quote_chars is None else set(quote_chars)
    parts = []
    temp_chars = []
    inside_quotes: str = None

    def _add_part():
        """ """
        nonlocal parts
        nonlocal temp_chars
        part = ''.join(temp_chars)
        if strip_parts is True:
            part = part.strip()
        for quote_char in quote_chars:
            if part.startswith(quote_char) and part.endswith(quote_char):
                part = part.strip(quote_char)
        if part:
            parts.append(part)
        temp_chars.clear()

    for char in text:
        if char in split_chars and inside_quotes is None:
            _add_part()
        else:
            temp_chars.append(char)

            if char in quote_chars and inside_quotes is None:
                inside_quotes = char
            elif char in quote_chars and inside_quotes == char:
                inside_quotes = None

    if temp_chars:
        _add_part()

    return parts


def make_attribute_name(in_string: str) -> str:
    """

    :param in_string: str:

    """

    # Remove invalid characters
    in_string = re.sub(r'-', '_', in_string)
    in_string = re.sub('[^0-9a-zA-Z_]', '', in_string)

    # Remove leading characters until we find a letter or underscore
    in_string = re.sub('^[^a-zA-Z_]+', '', in_string)

    return in_string.casefold()

# [2,"""",""__SERVER__"",false,2,""2""]


FIX_MULTIPLE_QUOTES_PATTERN = re.compile(r"""(\"|\')(\"{1,1}|\'{1,1})""")


def fix_multiple_quotes(_text: str) -> str:
    """

    :param _text: str:
    :param max_consecutive_quotes: int:  (Default value = None)

    """

    return FIX_MULTIPLE_QUOTES_PATTERN.sub("\\g<1>", _text)


def escape_doubled_quotes(text: str) -> str:
    """

    :param text: str:

    """
    def _replace_function(match: re.Match):
        """

        :param match: re.Match:

        """
        return r"\ ".strip() + match.group()[0]

    return re.sub(r"""(\"{2})|(\'{2})""", _replace_function, text)


def deindent(in_text: str, ignore_first_line: bool = False) -> str:
    """

    :param in_text: str:
    :param ignore_first_line: bool:  (Default value = False)

    """
    if in_text == "":
        return in_text
    pre_whitespace_regex = re.compile(r"\s*")
    lines = in_text.splitlines()
    white_space_levels = []
    if ignore_first_line is True:
        _first_line = lines.pop(0)
    for line in lines:
        if not line:
            continue
        if match := pre_whitespace_regex.match(line):
            ws = match.group()
            if len(ws) == len(line):
                continue
            white_space_levels.append(len(match.group()))
        else:

            white_space_levels.append(0)

    try:
        min_ws_level = min(white_space_levels) if len(white_space_levels) > 1 else white_space_levels[0]
    except IndexError:
        min_ws_level = 0
    combined = '\n'.join(line[min_ws_level:] for line in lines)
    if ignore_first_line is True:
        combined = '\n' + combined if combined else ""
        combined = _first_line + combined
    return combined


def multi_line_dedent(in_text: str, strip_pre_lines: bool = True, strip_post_lines: bool = True) -> str:
    """

    :param in_text: str:
    :param strip_pre_lines: bool:  (Default value = True)
    :param strip_post_lines: bool:  (Default value = True)

    """
    text = dedent(in_text)

    if strip_pre_lines is True:
        lines = text.splitlines()
        while lines[0] == "":
            lines.pop(0)
        text = '\n'.join(lines)
    if strip_post_lines is True:
        text = text.rstrip()
    return text


def strip_only_wrapping_empty_lines(in_text: str) -> str:
    """

    :param in_text: str:

    """
    empty_line_pattern = re.compile(r"(^\s*)|(\s*$)")
    return empty_line_pattern.sub("", in_text)


def string_strip(in_string: str, chars: str = None) -> str:
    """

    :param in_string: str:
    :param chars: str:  (Default value = None)

    """
    return in_string.strip(chars)


def remove_chars(in_string: str, *chars) -> str:
    """

    :param in_string: str:
    :param chars: Iterable[str]:

    """
    return ''.join(char for char in in_string if char not in set(chars))


def string_map_replace(in_string: str, replacement_map: Mapping[str, str]) -> str:
    new_string = str(in_string)
    for k, v in replacement_map.items():
        new_string = new_string.replace(k, v)

    return new_string


class RegexMapReplacer:
    """
    3x slower than just replace loop
    """
    __slots__ = ("_replacement_map", "_pattern")

    @mark_experimental()
    def __init__(self, replacement_map: Mapping[str, str]) -> None:
        self._replacement_map = replacement_map
        self._pattern: re.Pattern = re.compile(r"|".join(rf"{k}" for k in self._replacement_map))

    def _replacement_lookup(self, match: re.Match) -> str:
        orig_text_part = match.group()
        return self._replacement_map[orig_text_part]

    def apply(self, text: str) -> str:
        return self._pattern.sub(self._replacement_lookup, text)

    def __call__(self, text: str) -> Any:
        return self.apply(text=text)

# region [Main_Exec]


if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import re
from string import punctuation
from typing import TYPE_CHECKING, Any, Union, Callable, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QAction, QPixmap
from PySide6.QtCore import Qt, QSize, QObject
from PySide6.QtWidgets import QMenu, QFrame, QLabel, QHBoxLayout, QApplication, QWidgetAction, QSystemTrayIcon

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ApplicationNotExistingError
from gidapptools.general_helper.enums import MiscEnum

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.basics.application import GidQtApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class MenuTitle(QFrame):

    def __init__(self, text: str = None, icon: Union[QIcon, QPixmap] = None, parent=None):
        super().__init__(parent=parent)
        self.setup_styling()

        self.setLayout(QHBoxLayout())
        self.layout.setContentsMargins(5, 5, 5, 5)

        self.text_label = QLabel()
        if text:
            self.set_text(text)
        self.text_label.setAlignment(Qt.AlignCenter)

        self.icon_label = QLabel()

        if icon:
            self.set_icon(icon)
        self.icon_label.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(self.icon_label)
        self.layout.addWidget(self.text_label)
        self.widget_action = QWidgetAction(self)
        self.widget_action.setDefaultWidget(self)

    def setup_styling(self):
        self.setVisible(False)

        self.setFrameShape(QFrame.Panel)
        self.setFrameShadow(QFrame.Sunken)
        self.setMidLineWidth(1)
        self.setLineWidth(1)
        font = self.font()

        font.setPointSize(int(font.pointSize() * 0.9))
        font.setBold(True)

        font.setLetterSpacing(font.AbsoluteSpacing, 2)
        self.setFont(font)

    @property
    def layout(self) -> QHBoxLayout:
        return super().layout()

    @property
    def icon_size(self) -> QSize:
        if self.text_label.text():
            fm = self.fontMetrics()
            fm.horizontalAdvance(self.text_label.text())
            size = int(fm.height() * 0.9)
            return QSize(size, size)
        return QSize(20, 20)

    def set_text(self, text: str):
        self.text_label.setText(text)
        self.setVisible(True)

    def set_icon(self, icon: Union[QIcon, QPixmap]):
        if isinstance(icon, QIcon):
            icon = icon.pixmap(self.icon_size)

        else:
            icon = icon.scaled(self.icon_size, Qt.KeepAspectRatioByExpanding, Qt.SmoothTransformation)
        self.icon_label.setPixmap(icon)
        self.setVisible(True)


class GidBaseSysTray(QSystemTrayIcon):
    name_normalization_regex = re.compile(rf"[{punctuation}\s]+")

    def __init__(self, icon: Union[QIcon, QPixmap], title: str = None, tooltip: str = None, parent: Optional[QObject] = None):
        super().__init__(icon, parent=parent)
        self.setToolTip(tooltip)
        self.menu: QMenu = QMenu()
        self.setContextMenu(self.menu)
        self.menu_title = MenuTitle(title)
        self.actions: dict[str, QAction] = {}

    @property
    def app(self) -> "GidQtApplication":
        return QApplication.instance()

    @property
    def title(self) -> Optional[str]:
        return self.menu_title.text()

    @title.setter
    def title(self, value: str) -> None:
        self.menu_title.set_text(value)

    def setup(self) -> "GidBaseSysTray":
        if self.app is None:
            raise ApplicationNotExistingError(f"Setting up {self} before the application has been instantiated is forbidden.")
        self.menu.addAction(self.menu_title.widget_action)
        self.menu.addSeparator()

        self.setup_custom_actions()
        self.setup_standard_actions()
        self.menu.repaint()
        return self

    def setup_custom_actions(self):
        pass

    def setup_standard_actions(self):
        self.add_action("Close", self.app.quit)

    def _normalize_action_name(self, name: str) -> str:
        name = name.strip().casefold()
        name = self.name_normalization_regex.sub("_", name)
        return name

    def add_action(self, text: str, connect_to: Callable = None, icon: QIcon = None, enabled: bool = True, tooltip: str = None):
        name = self._normalize_action_name(text)
        if name in self.actions:
            raise KeyError(f"An action with the name {text!r} or the normalized name {name!r} already exists.")
        action = QAction(text, self.menu)
        action.setIconText(text)
        if icon:
            action.setIcon(icon)
        if tooltip:
            action.setToolTip(tooltip)
        action.setEnabled(enabled)

        self.actions[name] = action
        self.menu.addAction(action)
        if connect_to:
            action.triggered.connect(connect_to)

    def remove_action(self, name: str, error_on_missing: bool = False):
        if error_on_missing is True:
            action = self[name]
        else:
            action = self.get(name, MiscEnum.NOT_FOUND)
            if action is MiscEnum.NOT_FOUND:
                return

        self.menu.removeAction(action)
        self.actions.pop(self._normalize_action_name(name))

    def clear(self):
        for action_name in list(self.actions):
            self.remove_action(action_name)

    def __getitem__(self, key: str) -> QAction:
        try:
            return self.actions[self._normalize_action_name(key)]
        except KeyError as error:
            raise KeyError(f"No action named {key!r}") from error

    def get(self, key: str, default: Any = None) -> Union[QAction, Any]:
        return self.actions.get(key, default)
# region [Main_Exec]


if __name__ == '__main__':
    pass
    # endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Union, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QIcon, QAction, QPixmap
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QMenu, QStyle, QWidget, QMenuBar, QApplication

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import ApplicationNotExistingError

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gidapptools_qt.basics.application import GidQtApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseMenuBar(QMenuBar):
    general_disabled_action_names: frozenset[str] = frozenset()

    def __init__(self, parent: Optional[QWidget] = None, auto_connect_standard_actions: bool = True) -> None:
        super().__init__(parent=parent)
        self.auto_connect_standard_actions = auto_connect_standard_actions
        self.menus: list[QMenu] = []
        self.all_actions: dict[str, QAction] = {}

    def setup(self) -> "BaseMenuBar":
        self.setContextMenuPolicy(Qt.NoContextMenu)
        self.setup_default_menus()
        self.setup_extra_menus()

        self.handle_temp_disabled()

    def handle_temp_disabled(self) -> None:
        for action_name in self.general_disabled_action_names:
            action = self.all_actions.get(action_name)
            if action is not None:
                action.setEnabled(False)

    def setup_extra_menus(self):
        pass

    def setup_default_menus(self) -> None:
        self.file_menu = self.add_new_menu("File")
        self.exit_action = self.add_new_action(self.file_menu, "Exit")
        self.edit_menu = self.add_new_menu("Edit")

        self.view_menu = self.add_new_menu("View")
        self.windows_menu = self.add_new_menu("Windows", parent_menu=self.view_menu)

        self.settings_menu = self.add_new_menu("Settings")

        self.help_menu = self.add_new_menu("Help")
        self.help_menu.addSeparator()
        self.help_separator = self.help_menu.addSeparator()
        self.add_action(self.help_menu, self.help_separator)
        self.about_action = self.add_new_action(self.help_menu, "About", icon=self.style().standardIcon(QStyle.SP_MessageBoxInformation))
        self.about_action.setMenuRole(QAction.AboutRole)

        self.about_qt_action = self.add_new_action(self.help_menu, "About Qt®", icon=self.style().standardIcon(QStyle.SP_TitleBarMenuButton))
        self.about_qt_action.setMenuRole(QAction.AboutQtRole)
        if self.auto_connect_standard_actions is True:
            self.about_action.triggered.connect(self.app.show_about)
            self.about_qt_action.triggered.connect(self.app.show_about_qt)
            if self.parent() is not None:
                self.exit_action.triggered.connect(self.parent().close)

    @property
    def app(self) -> "GidQtApplication":
        app = QApplication.instance()
        if app is None:
            raise ApplicationNotExistingError()
        return app

    def add_new_menu(self, menu_title: str, icon: QIcon = None, add_before=None, parent_menu: QMenu = None) -> QMenu:
        menu = QMenu(self)
        menu.setEnabled(False)
        menu.setTitle(menu_title)
        if icon is not None:
            menu.setIcon(icon)

        target_menu = parent_menu or self
        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            target_menu.insertMenu(add_before, menu)
        else:
            target_menu.addMenu(menu)

        target_menu.setEnabled(True)
        return menu

    def add_new_action(self, menu: QMenu, action_name: str, action_title: str = None, icon: Union[QPixmap, QIcon] = None, add_before=None, action_klass: type[QAction] = QAction):
        action_title = action_title or action_name
        action_name = action_name.casefold().replace(' ', '_')
        action = action_klass(parent=menu)
        if icon:
            action.setIcon(icon)

        action.setText(action_title)
        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            menu.insertAction(add_before, action)
        else:
            menu.addAction(action)

        if not hasattr(menu, action_name + '_action'):
            setattr(menu, action_name + '_action', action)
        menu.setEnabled(True)
        self.all_actions[action_name] = action
        return action

    def add_action(self, menu: QMenu, action: QAction, add_before=None):

        if add_before is not None:
            if isinstance(add_before, QMenu):
                add_before = add_before.menuAction()
            menu.insertAction(add_before, action)
        else:
            menu.addAction(action)

        if not hasattr(menu, action.text() + '_action'):
            setattr(menu, action.text() + '_action', action)
        menu.setEnabled(True)
        return action

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(auto_connect_standard_actions={self.auto_connect_standard_actions!r})"


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
import PySide6
from PySide6.QtWidgets import QWidget, QApplication

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    pass

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ApplicationInfoWidget(QWidget):

    def __init__(self, app: QApplication, parent: Optional[PySide6.QtWidgets.QWidget] = None) -> None:
        self.app = app
        super().__init__(parent)

        # region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QFont
from PySide6.QtCore import Qt, QSize
from PySide6.QtWidgets import QLabel, QDialog, QLineEdit, QVBoxLayout

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class AppInfoDialog(QDialog):

    def __init__(self) -> None:
        super().__init__()

        self.label_font = self.create_label_font()
        self.parts = []
        self.setup()

    def create_label_font(self) -> QFont:
        font = QFont()
        font.setPointSize(19)
        font.setBold(True)
        return font

    def create_new_part(self, label_text: str, data_text: str) -> None:
        label = QLabel(self)
        label.setText(label_text)
        label.setFont(self.label_font)
        label.setAlignment(Qt.AlignLeft)

        data = QLineEdit(self)
        data.setReadOnly(True)
        data.setText(data_text)
        data.setAlignment(Qt.AlignRight)

        self.parts.append((label, data))

    def setup(self):
        self.resize(400, 100)
        self.setMaximumSize(QSize(400, 100))
        self.verticalLayout = QVBoxLayout(self)

        for label, data in self.parts:
            self.verticalLayout.addWidget(label)
            self.verticalLayout.addWidget(data)


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



from pathlib import Path
import os
from functools import cache
IMAGES_DIR = Path(__file__).parent.absolute()


class StoredImage:
    allowed_extensions: frozenset[str] = frozenset(["png", "jpg", "jpeg"])

    def __init__(self, path: Path):
        self.path = path.resolve()
        self.name = self.path.stem
        self.image_format = self.path.suffix.removeprefix('.')
        self._bytes: bytes = None

    @property
    def bytes(self) -> bytes:
        if self._bytes is None:
            self._bytes = self.path.read_bytes()
        return self._bytes


PLACEHOLDER_IMAGE = StoredImage(IMAGES_DIR.joinpath("placeholder.png"))
DEFAULT_APP_ICON_IMAGE = StoredImage(IMAGES_DIR.joinpath("default_app_icon.png"))


@cache
def get_image(name: str) -> StoredImage:
    cleaned_name = name.rsplit(".", 1)[0]
    for dirname, folderlist, filelist in os.walk(IMAGES_DIR):
        for file in filelist:
            if file.rsplit(".", 1)[-1] in StoredImage.allowed_extensions and file.casefold().rsplit(".", 1)[0] == cleaned_name:
                path = Path(dirname, file)
                image = StoredImage(path)

                return image
    raise FileNotFoundError(f"No image with name {name!r} found.")



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import Any, Optional
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import AppNameMissingError
from gidapptools.utility.enums import EnvName, NamedMetaPath
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.meta_data.config_kwargs import ConfigKwargs
from gidapptools.meta_data.meta_paths.meta_paths_item import MetaPaths
from gidapptools.abstract_classes.abstract_meta_factory import AbstractMetaFactory
from gidapptools.meta_data.meta_paths.appdirs_implementations import GidAppDirs

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class MetaPathsFactory(AbstractMetaFactory):
    appdirs_class = GidAppDirs
    product_class = MetaPaths
    default_configuration = {"roaming": MiscEnum.OPTIONAL, "multipath": MiscEnum.OPTIONAL}

    def __init__(self, config_kwargs: ConfigKwargs) -> None:
        super().__init__(config_kwargs=config_kwargs)
        self.code_base_dir = Path(self.config_kwargs.get('init_path')).parent

    def get_path_dict(self) -> dict[NamedMetaPath, Optional[Path]]:
        defaults = {'app_name': os.getenv(EnvName.APP_NAME.value),
                    'app_author': os.getenv(EnvName.APP_AUTHOR.value)}
        _kwargs = self.config_kwargs.get_kwargs_for(self.appdirs_class.get_path_dict_direct, defaults)
        if _kwargs.get('app_name') is None:
            raise AppNameMissingError()

        path_overwrites = self.config_kwargs.get('path_overwrites', {})
        if self.config_kwargs.created_meta_items["meta_info"].is_dev is True and NamedMetaPath.DEBUG_DUMP not in path_overwrites:
            path_overwrites[NamedMetaPath.DEBUG_DUMP] = self.code_base_dir.joinpath("debug_dump")
        path_dict = self.appdirs_class.get_path_dict_direct(**_kwargs)
        return path_dict | path_overwrites

    @classmethod
    @property
    def __default_configuration__(cls) -> dict[str, Any]:
        return cls.default_configuration | cls.product_class.__default_configuration__

    def setup(self) -> None:
        self.path_dict = self.get_path_dict()
        self.is_setup = True

    def _build(self) -> None:
        if self.is_setup is False:
            self.setup()
        _kwargs = self.config_kwargs.get_kwargs_for(self.product_class, defaults={'code_base_dir': self.code_base_dir, 'paths': self.path_dict})
        instance = self.product_class(**_kwargs)

        return instance


# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import asyncio
import inspect
from abc import ABC, abstractmethod
from typing import Any, Union, Callable, Hashable
from pathlib import Path
from weakref import WeakSet, WeakMethod, ref

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.helper import get_qualname_or_name

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class AdaptableWeakSet(WeakSet):

    def add(self, item) -> None:
        if self._pending_removals:
            self._commit_removals()

        elif inspect.ismethod(item):
            ref_item = WeakMethod(item, self._remove)
        else:
            ref_item = ref(item, self._remove)
        self.data.add(ref_item)

    def __bool__(self) -> bool:
        return self.data is not None and len(self.data) > 0


class AbstractSignal(ABC):

    def __init__(self, key: Hashable, allow_sync_targets: bool = True, allow_async_target: bool = True) -> None:
        self.key = key
        self.allow_sync_targets = allow_sync_targets
        self.allow_async_targets = allow_async_target
        if self.allow_async_targets is False and self.allow_async_targets is False:
            raise AttributeError('A signal cannot have both SYNC and ASYNC targets disabled.')
        self.targets = AdaptableWeakSet()
        self.targets_info: dict[str, dict[str:Any]] = {}

    def _add_target_info(self, target: Callable) -> None:
        info = {'is_coroutine': asyncio.iscoroutine(target)}
        name = get_qualname_or_name(target)
        self.targets_info[name] = info

    def _verify(self, target: Callable) -> None:
        if inspect.isbuiltin(target):
            raise TypeError('cannot weakreference built_ins.')
        target_is_coroutine = asyncio.iscoroutine(target)
        if target_is_coroutine is True and self.allow_async_targets is False:
            raise TypeError('Signal is set to only allow SYNC targets.')
        if target_is_coroutine is False and self.allow_sync_targets is False:
            raise TypeError('Signal is set to only allow ASYNC targets.')

    def connect(self, target: Callable) -> None:
        self._verify(target)
        self._add_target_info(target)
        self.targets.add(target)

    def disconnect(self, target: Callable) -> None:
        name = get_qualname_or_name(target)
        self.targets.discard(target)
        self.targets_info.pop(name)

    @abstractmethod
    def fire_and_forget(self, *args, **kwargs):
        ...

    @abstractmethod
    def delayed_fire_and_forget(self, delay: Union[int, float], *args, **kwargs):
        ...

    @abstractmethod
    def emit(self, *args, **kwargs):
        ...

    @abstractmethod
    async def aemit(self, *args, **kwargs):
        ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(key={self.key!r})"

    def __str__(self):
        return f"{self.__class__.__name__}-{self.key}(targets={self.targets!r})"


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
from typing import Any, Union, Literal, Callable, Optional
from pathlib import Path
from datetime import datetime

# * Third Party Imports --------------------------------------------------------------------------------->
from tzlocal import get_localzone
from rich.box import SQUARE
from rich.text import Text as RichText
from rich.panel import Panel
from rich.style import Style
from rich.console import Console as RichConsole
from rich.console import JustifyMethod
from rich.logging import LogRender
from rich.measure import measure_renderables
from rich.protocol import is_renderable
from rich.containers import Renderables

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

CONSOLE = RichConsole(soft_wrap=True)


def dprint(*args, **kwargs):
    CONSOLE.print(*args, **kwargs)
    CONSOLE.rule()


class FakeLogger:
    styled_log_levels: dict[str, RichText] = {"debug": RichText.styled('DEBUG'.center(8), "logging.level.debug"),
                                              "info": RichText.styled('INFO'.center(8), "logging.level.info"),
                                              "warning": RichText.styled('WARNING'.center(8), "logging.level.warning"),
                                              "error": RichText.styled('ERROR'.center(8), "logging.level.error"),
                                              "critical": RichText.styled('CRITICAL'.center(8), "logging.level.critical")}

    def __init__(self,
                 default_sep: str = " ",
                 default_end: str = "\n",
                 default_style: Optional[Union[str, Style]] = None,
                 default_justify: JustifyMethod = None,
                 as_panel: bool = False,
                 dynamic_panel_title: Union[Literal["output_number", "datetime", "level", "all"], Callable] = "level",
                 ** console_kwargs) -> None:
        self.default_sep = default_sep
        self.default_end = default_end
        self.default_style = default_style
        self.default_justify = default_justify
        _ = console_kwargs.pop("soft_wrap", None)
        self.as_panel = as_panel
        self.dynamic_panel_title = dynamic_panel_title
        self.console = RichConsole(soft_wrap=True, **console_kwargs)
        self.console._log_render = LogRender(show_time=True, show_level=True, omit_repeated_times=False)
        self._output_number = 0
        self._dynamic_panel_title_table = {"output_number": self._get_output_number,
                                           "datetime": self._get_datetime,
                                           "level": self._get_level,
                                           "all": self._get_all,
                                           None: lambda: None}
        self.console.clear()

    def _get_all(self) -> str:
        return '  ' + ' | '.join((self._get_datetime(), self._get_level(extra_back=True), self._get_output_number())) + '  '

    def _get_level(self, extra_back: bool = False) -> str:
        if extra_back is True:
            caller_name = inspect.currentframe().f_back.f_back.f_back.f_code.co_name
        else:
            caller_name = inspect.currentframe().f_back.f_back.f_code.co_name
        return caller_name.upper()

    def _get_datetime(self) -> str:
        date_time = datetime.now(tz=get_localzone())
        return date_time.strftime("%x %X")

    def _get_output_number(self) -> str:
        return str(self._output_number)

    def _get_rule_title(self) -> str:
        if callable(self.dynamic_panel_title):
            return self.dynamic_panel_title()
        return self._dynamic_panel_title_table.get(self.dynamic_panel_title)()

    def debug(self,
              *objects: Any,
              sep: str = ...,
              end: str = ...,
              style: Optional[Union[str, Style]] = ...,
              justify: Optional[JustifyMethod] = ...,
              emoji: Optional[bool] = None,
              markup: Optional[bool] = None,
              highlight: Optional[bool] = None) -> None:
        sep = self.default_sep if sep is ... else sep
        end = self.default_end if end is ... else end
        style = self.default_style if style is ... else style
        justify = self.default_justify if justify is ... else justify
        self._output_number += 1
        objects = tuple(self._fix_obj(o) for o in objects)
        if self.as_panel is True:
            objects = [Panel(Renderables(objects), title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="grey66", color="grey3")), box=SQUARE)]

        self.console.rule(title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="grey66", color="grey3")))
        self.console.print(*objects, sep=sep, end=end, style=style, justify=justify, emoji=emoji, markup=markup, highlight=highlight)

    def info(self,
             *objects: Any,
             sep: str = ...,
             end: str = ...,
             style: Optional[Union[str, Style]] = ...,
             justify: Optional[JustifyMethod] = ...,
             emoji: Optional[bool] = None,
             markup: Optional[bool] = None,
             highlight: Optional[bool] = None) -> None:
        sep = self.default_sep if sep is ... else sep
        end = self.default_end if end is ... else end
        style = self.default_style if style is ... else style
        justify = self.default_justify if justify is ... else justify
        self._output_number += 1
        objects = tuple(self._fix_obj(o) for o in objects)
        if self.as_panel is True:
            objects = [Panel(Renderables(objects), title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="light_sky_blue1", color="dark_green")), box=SQUARE)]

        self.console.rule(title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="light_sky_blue1", color="dark_green")))
        self.console.print(*objects, sep=sep, end=end, style=style, justify=justify, emoji=emoji, markup=markup, highlight=highlight)

    def warning(self,
                *objects: Any,
                sep: str = ...,
                end: str = ...,
                style: Optional[Union[str, Style]] = ...,
                justify: Optional[JustifyMethod] = ...,
                emoji: Optional[bool] = None,
                markup: Optional[bool] = None,
                highlight: Optional[bool] = None) -> None:
        sep = self.default_sep if sep is ... else sep
        end = self.default_end if end is ... else end
        style = self.default_style if style is ... else style
        justify = self.default_justify if justify is ... else justify
        self._output_number += 1
        objects = tuple(self._fix_obj(o) for o in objects)
        if self.as_panel is True:
            objects = [Panel(Renderables(objects), title=RichText.styled(self._get_rule_title(), style=Style(overline=True, underline=True, color="bright_white")), box=SQUARE)]

        self.console.rule(title=RichText.styled(self._get_rule_title(), style=Style(overline=True, underline=True, color="bright_white")))
        self.console.print(*objects, sep=sep, end=end, style=style, justify=justify, emoji=emoji, markup=markup, highlight=highlight)

    def critical(self,
                 *objects: Any,
                 sep: str = ...,
                 end: str = ...,
                 style: Optional[Union[str, Style]] = ...,
                 justify: Optional[JustifyMethod] = ...,
                 emoji: Optional[bool] = None,
                 markup: Optional[bool] = None,
                 highlight: Optional[bool] = None) -> None:
        sep = self.default_sep if sep is ... else sep
        end = self.default_end if end is ... else end
        style = self.default_style if style is ... else style
        justify = self.default_justify if justify is ... else justify
        self._output_number += 1
        objects = tuple(self._fix_obj(o) for o in objects)
        if self.as_panel is True:
            objects = [Panel(Renderables(objects), title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="dark_red", color="white")), box=SQUARE)]

        width = measure_renderables(self.console, options=self.console.options, renderables=objects).maximum
        self.console.line()
        self.console.rule(title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="dark_red", color="white")))
        self.console.print("!" * width)
        self.console.print(*objects, sep=sep, end=end, style=style, justify=justify, emoji=emoji, markup=markup, highlight=highlight)
        self.console.print("!" * width)
        self.console.line()

    def error(self,
              *objects: Any,
              sep: str = ...,
              end: str = ...,
              style: Optional[Union[str, Style]] = ...,
              justify: Optional[JustifyMethod] = ...,
              emoji: Optional[bool] = None,
              markup: Optional[bool] = None,
              highlight: Optional[bool] = None) -> None:
        sep = self.default_sep if sep is ... else sep
        end = self.default_end if end is ... else end
        style = self.default_style if style is ... else style
        justify = self.default_justify if justify is ... else justify
        self._output_number += 1
        objects = tuple(self._fix_obj(o) for o in objects)

        if self.as_panel is True:
            objects = [Panel(Renderables(objects), title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="yellow3", color="red")), box=SQUARE)]

        self.console.rule(title=RichText.styled(self._get_rule_title(), style=Style(bold=True, bgcolor="yellow3", color="red")))
        self.console.print(*objects, sep=sep, end=end, style=style, justify=justify, emoji=emoji, markup=markup, highlight=highlight)

    @staticmethod
    def _fix_obj(in_obj):
        if is_renderable(in_obj):
            return in_obj
        return str(in_obj)


fake_logger = FakeLogger(as_panel=False, dynamic_panel_title="all", default_justify=None)

# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
import inspect
from pprint import pprint
from typing import TYPE_CHECKING, Any, Union
from pathlib import Path
from collections.abc import Mapping, Sequence

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# * Standard Library Imports ---------------------------------------------------------------------------->
import argparse

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.import_helper import import_from_name, meta_data_from_module

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def url_sorter(in_name_and_url: tuple[str, str]):
    order = ("homepage",
             "home-page",
             "documentation",
             "wiki",
             "changelog",
             "changes"
             "source",
             "github",
             "repository",
             "source code",
             "source",
             "download-url")
    name = in_name_and_url[0]
    try:

        return order.index(name.casefold())

    except ValueError:
        return 999


class GidArgumentParser(argparse.ArgumentParser):

    def __init__(self,
                 prog: str = None,
                 usage: str = None,
                 description: str = None,
                 epilog: str = None,
                 parents: Sequence[argparse.ArgumentParser] = None,
                 formatter_class: "argparse._FormatterClass" = argparse.HelpFormatter,
                 prefix_chars: str = '-',
                 fromfile_prefix_chars: str = None,
                 argument_default: Any = None,
                 conflict_handler: str = 'error',
                 add_help: bool = True,
                 allow_abbrev: bool = True,
                 exit_on_error: bool = True,
                 version: str = None,
                 urls: Union[str, Mapping[str, str]] = None) -> None:

        super().__init__(prog=prog,
                         usage=usage,
                         description=description,
                         epilog=epilog,
                         parents=parents or [],
                         formatter_class=formatter_class,
                         prefix_chars=prefix_chars,
                         fromfile_prefix_chars=fromfile_prefix_chars,
                         argument_default=argument_default,
                         conflict_handler=conflict_handler,
                         add_help=add_help,
                         allow_abbrev=allow_abbrev,
                         exit_on_error=exit_on_error)

        self.version = version
        self._urls = urls
        self.standard_arguments: dict[str, argparse.Action] = {}

    @property
    def urls(self) -> dict[str, str]:
        if self._urls is None:
            return {}
        if isinstance(self._urls, str):
            return {"Homepage", self._urls}

        return dict(self._urls)

    def setup_standard_arguments(self) -> Self:
        if self.version:
            self._add_to_standard_arguments(self.add_argument("-v", "--version", action=argparse._VersionAction))

        return self

    def _add_to_standard_arguments(self, action: argparse.Action):
        for opt_string in action.option_strings:
            name = opt_string.strip(self.prefix_chars)
            self.standard_arguments[name] = action

    def format_help(self):
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)
        for url_name, url in sorted(self.urls.items(), key=url_sorter):
            formatter.add_text(f"{url_name}: {url}")

        # positionals, optionals and user-defined groups
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        if self.epilog:
            formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()

    @classmethod
    def from_meta_data(cls, package_name: str, pretty_name: str = None, **kwargs) -> "GidArgumentParser":
        package_module = import_from_name(package_name)
        meta_data = meta_data_from_module(package_module)
        init_kwargs = {}
        init_kwargs["version"] = getattr(package_module, "__version__", meta_data["version"])
        init_kwargs["prog"] = pretty_name or meta_data["name"]
        init_kwargs["description"] = meta_data.get("summary") or (inspect.getdoc(package_module) or "")
        init_kwargs["urls"] = meta_data.all_urls

        init_kwargs = init_kwargs | kwargs
        return cls(**init_kwargs)
# region [Main_Exec]


if __name__ == '__main__':
    x = import_from_name("gidapptools")
    y = meta_data_from_module(x)
    pprint(y)

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Optional
from pathlib import Path

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtWidgets import QLayout, QWidget

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class VerticalFormLayout(QLayout):

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP
"""

from gidapptools.meta_data.interface import setup_meta_data, get_meta_info, get_meta_item, get_meta_paths
from gidapptools.gid_logger.logger import setup_main_logger, get_logger, setup_main_logger_with_file_logging, get_main_logger, get_handlers


__version__ = "0.5.7"


from pathlib import Path
from tzlocal import reload_localzone

THIS_FILE_DIR = Path(__file__).resolve().parent

# _log = logging.getLogger(__name__)
# _log.addHandler(logging.NullHandler())


def setup_library() -> None:
    from .gid_logger.logger import make_library_logger
    log = make_library_logger(__name__)
    log.debug("library %r setup", __name__)

    reload_localzone()


setup_library()



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Optional
from pathlib import Path
from tempfile import gettempdir

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.utility.helper import PathLibAppDirs

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class GidAppDirs(PathLibAppDirs):

    @PathLibAppDirs.mark_path
    def user_config_dir(self) -> Path:
        config_dir = super().user_config_dir().joinpath('config')
        return config_dir

    @PathLibAppDirs.mark_path
    def user_config_spec_dir(self) -> Path:
        config_spec_dir = self.user_config_dir().joinpath('_spec')
        return config_spec_dir

    @PathLibAppDirs.mark_path
    def user_cache_dir(self) -> Path:
        cache_dir = super().user_cache_dir()
        return cache_dir.with_name(cache_dir.name.lower())

    @PathLibAppDirs.mark_path
    def user_log_dir(self) -> Path:
        log_dir = super().user_log_dir()
        return log_dir.with_name(log_dir.name.lower())

    @PathLibAppDirs.mark_path
    def user_temp_dir(self) -> Path:
        default_temp_dir = Path(gettempdir())
        app_temp_dir = default_temp_dir.joinpath(self.appname)
        return Path(str(app_temp_dir))

    @PathLibAppDirs.mark_path
    def database_dir(self) -> Path:
        return self.user_data_dir().joinpath('database')

    @PathLibAppDirs.mark_path
    def debug_dump_dir(self) -> Path:
        return self.user_log_dir().joinpath("debug_dump")

    @classmethod
    def get_path_dict_direct(cls,
                             app_name: str,
                             app_author: str = None,
                             roaming: bool = True,
                             multipath: bool = False) -> dict[NamedMetaPath, Optional[Path]]:
        inst = cls(appname=app_name, appauthor=app_author, roaming=roaming, multipath=multipath)
        return inst.as_path_dict()


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Any
from pathlib import Path
from threading import Event

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BlockingEvent(Event):

    def __init__(self) -> None:
        super().__init__()
        self.set()

    def activate_blocking(self) -> None:
        self.clear()

    def clear_blocking(self) -> None:
        self.set()

    def __enter__(self) -> None:
        self.clear()

    def __exit__(self, exception_type: type = None, exception_value: BaseException = None, traceback: Any = None) -> None:
        self.set()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(state={self.is_set()!r})'

    def wait(self, timeout: float | None = None) -> bool:
        return super().wait(timeout)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



from pathlib import Path


TEMPLATES_DIR = Path(__file__).parent.absolute()

ABOUT_STYLESHEET_FILE = TEMPLATES_DIR.joinpath("about_stylesheet.css")

ABOUT_TEMPLATE_FILE = TEMPLATES_DIR.joinpath("about_template.jinja_html")


ARG_DOC_HTML_TEMPLATE_FILE = TEMPLATES_DIR.joinpath("arg_doc_html_template.jinja_html")

ARG_DOC_MARKDOWN_TEMPLATE_FILE = TEMPLATES_DIR.joinpath("arg_doc_markdown_template.jinja_md")



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from rich.console import Console as RichConsole
from rich.style import Style
from rich.default_styles import DEFAULT_STYLES
# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

import os
import re
import sys
import json
import queue
import math
import base64
import pickle
import random
import shelve
import dataclasses
import shutil
import asyncio
import logging
import sqlite3
import platform
from math import sqrt
import subprocess
import inspect
from decimal import Decimal

from time import sleep, process_time, process_time_ns, perf_counter, perf_counter_ns
from io import BytesIO, StringIO
from abc import ABC, ABCMeta, abstractmethod
from copy import copy, deepcopy
from enum import Enum, Flag, auto, unique
from pprint import pprint, pformat
from pathlib import Path
from string import Formatter, digits, printable, whitespace, punctuation, ascii_letters, ascii_lowercase, ascii_uppercase
from timeit import Timer
from types import ModuleType
from typing import (TYPE_CHECKING, TypeVar, TypeGuard, TypeAlias, Final, TypedDict, Generic, Union, Optional, ForwardRef, final,
                    no_type_check, no_type_check_decorator, overload, get_type_hints, cast, Protocol, runtime_checkable, NoReturn, NewType, Literal, AnyStr, IO, BinaryIO, TextIO, Any)
from collections import Counter, ChainMap, deque, namedtuple, defaultdict
from collections.abc import (AsyncGenerator, AsyncIterable, AsyncIterator, Awaitable, ByteString, Callable, Collection, Container, Coroutine, Generator,
                             Hashable, ItemsView, Iterable, Iterator, KeysView, Mapping, MappingView, MutableMapping, MutableSequence, MutableSet, Reversible, Sequence, Set, Sized, ValuesView)
from zipfile import ZipFile, ZIP_LZMA
from datetime import datetime, timezone, timedelta
from tempfile import TemporaryDirectory
from textwrap import TextWrapper, fill, wrap, dedent, indent, shorten
from functools import wraps, partial, lru_cache, singledispatch, total_ordering, cached_property, cache
from contextlib import contextmanager, asynccontextmanager, nullcontext, closing, ExitStack, suppress
from statistics import mean, mode, stdev, median, variance, pvariance, harmonic_mean, median_grouped
from urllib.parse import urlparse
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, wait, as_completed, ALL_COMPLETED, FIRST_EXCEPTION, FIRST_COMPLETED

import PySide6
from PySide6 import (QtCore, QtGui, QtWidgets, Qt3DAnimation, Qt3DCore, Qt3DExtras, Qt3DInput, Qt3DLogic, Qt3DRender, QtAxContainer, QtBluetooth,
                     QtCharts, QtConcurrent, QtDataVisualization, QtDesigner, QtHelp, QtMultimedia, QtMultimediaWidgets, QtNetwork, QtNetworkAuth,
                     QtOpenGL, QtOpenGLWidgets, QtPositioning, QtPrintSupport, QtQml, QtQuick, QtQuickControls2, QtQuickWidgets, QtRemoteObjects,
                     QtScxml, QtSensors, QtSerialPort, QtSql, QtStateMachine, QtSvg, QtSvgWidgets, QtTest, QtUiTools, QtWebChannel, QtWebEngineCore,
                     QtWebEngineQuick, QtWebEngineWidgets, QtWebSockets, QtXml)

from PySide6.QtCore import (QByteArray, QCoreApplication, QDate, QDateTime, QEvent, QLocale, QMetaObject, QModelIndex, QModelRoleData, QMutex,
                            QMutexLocker, QObject, QPoint, QRect, QRecursiveMutex, QRunnable, QSettings, QSize, QThread, QThreadPool, QTime, QUrl,
                            QWaitCondition, Qt, QAbstractItemModel, QAbstractListModel, QAbstractTableModel, Signal, Slot)

from PySide6.QtGui import (QAction, QTextBlockFormat, QTextCursor, QBrush, QTextOption, QSyntaxHighlighter, QCloseEvent, QColor, QConicalGradient, QCursor, QFont, QFontDatabase, QFontMetrics, QGradient, QIcon, QImage,
                           QKeySequence, QLinearGradient, QPainter, QPalette, QPixmap, QRadialGradient, QTransform)

from PySide6.QtWidgets import (QApplication, QStyleOptionGroupBox, QScrollArea, QBoxLayout, QCheckBox, QColorDialog, QColumnView, QComboBox, QDateTimeEdit, QDialogButtonBox,
                               QDockWidget, QDoubleSpinBox, QFontComboBox, QFormLayout, QFrame, QGridLayout, QGroupBox, QHBoxLayout, QHeaderView,
                               QLCDNumber, QLabel, QLayout, QLineEdit, QListView, QListWidget, QMainWindow, QMenu, QMenuBar, QMessageBox,
                               QProgressBar, QProgressDialog, QPushButton, QSizePolicy, QSpacerItem, QSpinBox, QStackedLayout, QStackedWidget,
                               QStatusBar, QStyledItemDelegate, QSystemTrayIcon, QTabWidget, QTableView, QTextEdit, QTimeEdit, QToolBox, QTreeView,
                               QVBoxLayout, QWidget, QLayoutItem, QAbstractItemDelegate, QAbstractItemView, QAbstractScrollArea, QRadioButton, QFileDialog, QButtonGroup)

import shiboken6
from gidapptools.gidapptools_qt.widgets.separator_lines import HorizontalSeparatorLine, VerticalSeparatorLine
from gidapptools.general_helper.string_helper import shorten_string, StringCaseConverter, StringCase
from gidapptools.gidapptools_qt._data.images import get_image
from gidapptools.gidapptools_qt.layouts.sorted_layout import SortedVBoxLayout


import distinctipy
from pygments import lexers
from pygments.lexer import Lexer
from pygments import formatters
from pygments import filters
from pygments.filter import Filter
from pygments.formatter import Formatter
from pygments import highlight, token

from gidapptools.gidapptools_qt.debug.debug_value_widgets import determine_value_widget
if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class DebugContentWidget(QWidget):

    def __init__(self, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.setLayout(QVBoxLayout())

    @property
    def layout(self) -> QVBoxLayout:
        return super().layout()


class DebugDialog(QScrollArea):

    def __init__(self,
                 title: str,
                 key_text: str,
                 category: "DebugCategoryBox") -> None:
        super().__init__(parent=None)

        self.setWindowTitle(title)
        self.key_text = key_text
        self.value_data: object = None
        self.category = category
        self.content_widget = DebugContentWidget()

        self.key_text_widget: QTextEdit = self._make_key_text_widget()

        self.separator_line = HorizontalSeparatorLine(self)

        self._value_widget = None
        self.is_setup: bool = False
        # self._setup()

    @property
    def value_widget(self) -> QWidget:
        if self._value_widget is None:
            self._value_widget = self._make_value_widget()
        return self._value_widget

    def _make_key_text_widget(self) -> QTextEdit:

        key_text_widget = QLabel()
        key_text_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)
        font = key_text_widget.font()
        font.setBold(True)
        font.setPointSize(int(font.pointSize() * 1.5))
        key_text_widget.setFont(font)
        return key_text_widget

    def _make_value_widget(self) -> QWidget:
        value_widget = determine_value_widget(self.value_data)(self)
        try:
            value_widget = value_widget.setup()
        except AttributeError:
            pass
        return value_widget

    def _setup_scrollbar(self) -> None:
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setWidgetResizable(True)

    def _setup_widgets(self) -> None:
        self.content_widget.layout.insertWidget(0, self.key_text_widget, 0)
        self.content_widget.layout.insertWidget(1, self.separator_line, 0)

        self.key_text_widget.setText(self.key_text)
        self.key_text_widget.adjustSize()

    def setup(self) -> Self:
        if self.is_setup is True:
            return
        self.setWidget(self.content_widget)

        self._setup_widgets()
        self._setup_scrollbar()
        self.resize(1000, 600)

        self.is_setup = True
        return self

    def set_data(self, data: object):
        self.value_data = data
        self.content_widget.layout.insertWidget(2, self.value_widget, 1)

        self.content_widget.layout.addStretch()
        self.value_widget.set_data(self.value_data)
        # self.repaint()

    def show(self):

        super().show()

    def closeEvent(self, event: QCloseEvent) -> None:
        if event.isAccepted():
            self.category.result_window_closed.emit(self)

        return super().closeEvent(event)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(title={self.windowTitle()!r})"


class ShowAttrButton(QPushButton):

    def __init__(self, attr_name: str, obj: Union[object, type, ModuleType], parent=None):
        super().__init__(parent=parent)
        self.setObjectName("ShowAttrButton")
        self.attr_name = attr_name
        self.obj = obj
        if inspect.isclass(self.obj) or inspect.ismodule(self.obj):
            self.obj_name = obj.__name__
        else:
            self.obj_name = obj.__class__.__name__

        self.category_name = self.obj_name

        self.setText(f"Show {self.obj_name}.{self.attr_name}")

        self.clicked.connect(self.show_info_box)
        self.setStyleSheet("""

        ShowAttrButton {
            background-color: rgba(173, 216, 230, 34%);
        }

        """)

    def show_info_box(self):
        title = self.attr_name

        attr_value = getattr(self.obj, self.attr_name)
        if callable(attr_value):
            attr_value = attr_value()

        # key_text = f"Attribute <i><b>{self.attr_name!r}</b></i> of object <i><b>{self.obj!r}</b></i> is:"
        key_text = f"Attribute {self.attr_name!r} of object {self.obj!r} is:"

        dialog = DebugDialog(title=title, key_text=key_text, category=self.parent()).setup()
        dialog.set_data(attr_value)
        self.parent().result_window_created.emit(dialog)
        dialog.show()


class ShowFunctionResultButton(QPushButton):
    def __init__(self, category_name: str, function: Callable, parent=None, **kwargs):
        super().__init__(parent=parent)
        self.setObjectName("ShowFunctionResultButton")

        self.category_name = category_name
        self.function = function
        self.kwargs = kwargs
        self._text = self.function.__name__
        if kwargs:
            self._text += " with " + ', '.join(f"{k}=>{shorten_string(repr(v), 50, split_on=r'any')}" for k, v in self.kwargs.items())

        self.setText(f"show result for {self._text}")
        self.clicked.connect(self.show_info_box)
        if getattr(self.function, "is_disabled", False) is True:
            self.setEnabled(False)
        self.setStyleSheet("""

        ShowFunctionResultButton {
            background-color: rgba(153, 229, 153, 34%);
        }

        """)

    @property
    def app(self) -> Optional[QApplication]:
        return QApplication.instance()

    def show_info_box(self):
        title = f"Result for {self.function.__name__}"
        value = self.function(**self.kwargs)
        if inspect.isgenerator(value):
            value = tuple(value)
        # key_text = f"Result for <i><b>{self._text}</b></i> is:"
        key_text = f"Result for {self._text} is:"

        dialog = DebugDialog(title=title, key_text=key_text, category=self.parent()).setup()
        dialog.set_data(value)
        self.parent().result_window_created.emit(dialog)

        dialog.show()


class DebugCategoryBox(QGroupBox):
    result_window_created = Signal(QWidget)
    result_window_closed = Signal(QWidget)

    def __init__(self, name: str, parent=None):
        super().__init__(parent=parent)
        self.setObjectName("DebugCategoryBox")

        self.name = name
        self.setTitle(" " + self.name + ' ')
        self.setLayout(QVBoxLayout())
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.widgets: dict[str, QWidget] = {}
        self._mod_font()
        self.setCheckable(True)

        self.toggled.connect(self.on_toggled)

    def _mod_font(self) -> None:
        font = self.font()
        font.setBold(True)
        font.setPointSize(int(font.pointSize() * 1.25))
        self.setFont(font)

    def on_toggled(self, value: bool):
        for child_idx in range(self.layout.count()):
            child = self.layout.itemAt(child_idx)
            child.widget().setVisible(value)

    @ property
    def layout(self) -> QVBoxLayout:
        return super().layout()

    def add_widget(self, name: str, widget: QWidget):
        if name in self.widgets:
            raise ValueError(f"Name {name!r} already registered in category {self!r}.")
        self.layout.addWidget(widget)
        widget.setParent(self)
        self.widgets[name] = widget


def layout_sort_func(in_item: QLayoutItem):
    try:
        name = in_item.widget().name.casefold()
    except AttributeError:
        name = in_item.widget().objectName().casefold()

    general_pos = 0
    if name.endswith("-attribute"):
        general_pos = 1

    return general_pos, name.removeprefix("show ")


class DebugWidget(QDockWidget):

    def __init__(self, inject_app: bool = False, func_injects: dict[str, object] = None, parent: QWidget = None) -> None:
        super().__init__(parent)
        self.inject_app = inject_app
        self.func_injects = func_injects or {}
        self.scroll_area = QScrollArea()

        self.content_widget = QWidget(self)
        self.content_widget.setLayout(SortedVBoxLayout(layout_sort_func))
        # self.content_widget.setLayout(QVBoxLayout())
        # self.content_widget.layout().setContentsMargins(25, 25, 25, 25)
        self.setWindowTitle("Debug")
        self.setWindowIcon(QPixmap(get_image("debugging_icon.png").path))

        self.categories: dict[str, DebugCategoryBox] = {}
        self._current_result_window: QWidget = None

    @property
    def app(self) -> Optional[QApplication]:
        return QApplication.instance()

    def set_current_result_window(self, window: QWidget) -> None:
        if self._current_result_window is not None:
            self._current_result_window.close()

        self._current_result_window = window

    def unset_current_result_window(self, window: QWidget) -> None:
        if self._current_result_window is None:
            return

        elif self._current_result_window is not window:
            return

        else:

            self._current_result_window = None

    def setup(self) -> Self:
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(self.content_widget)

        self.scroll_area.verticalScrollBar().setSingleStep(2)

        self.setWidget(self.scroll_area)

        self.setMinimumSize(QSize(300, 300))

        self.resize(600, 600)
        return self

    def get_category_box(self, name: str) -> DebugCategoryBox:
        try:
            return self.categories[name.strip().casefold()]
        except KeyError:
            category = DebugCategoryBox(name=name)
            category.result_window_created.connect(self.set_current_result_window)
            category.result_window_closed.connect(self.unset_current_result_window)
            self.content_widget.layout().addWidget(category)
            self.categories[name.strip().casefold()] = category
            return category

    def _modify_func_globals(self, func: Callable) -> None:
        if not hasattr(func, "__globals__"):
            return
        for k, v in self.func_injects.items():
            if k not in func.__globals__:
                func.__globals__[k] = v

        if self.inject_app is True and "app" not in func.__globals__:
            func.__globals__["app"] = self.app

    def add_widget(self, name: str, category: str, widget: QWidget):
        category_box = self.get_category_box(category)
        category_box.add_widget(name=name, widget=widget)

    def add_show_attr_button(self, attr_name: str, obj: object):
        button = ShowAttrButton(attr_name=attr_name, obj=obj, parent=self)
        self.add_widget(button.text(), button.category_name, button)

    def add_show_func_result_button(self, function: Callable, category_name: str, **kwargs):
        self._modify_func_globals(func=function)
        button = ShowFunctionResultButton(category_name=category_name, function=function, **kwargs)
        self.add_widget(button.text(), button.category_name, button)

    def closeEvent(self, event: QCloseEvent) -> None:
        if event.isAccepted():
            if self._current_result_window is not None:

                self._current_result_window.close()
        super().closeEvent(event)

    def show(self) -> None:
        return super().show()
# region [Main_Exec]


def stupid_function():
    return "aaaaa", 13, 14.5, ["a", "list", "not", "a", "tuple"], {"dog": "cat", "woof": 1337}


def stupid_function_2():
    return "woof"


def stupid_function_3():
    return 3


def stupid_function_4():
    return sqrt(2)


def stupid_function_5():
    return Decimal(2).sqrt()


def get_all_lexers():
    _out = []
    for name, *_ in lexers.get_all_lexers():
        _out.append(name)

    return tuple(_out)


def inspect_filter(name: str):
    _filter: Filter = filters.get_filter_by_name(name)
    return {"name": _filter.__class__.__qualname__,
            "module": _filter.__module__,
            "options": _filter.options,
            "doc": _filter.__doc__,
            "dict": _filter.__dict__}


def inspect_all_filter():
    for filter_name in filters.get_all_filters():
        yield inspect_filter(filter_name)


if __name__ == '__main__':
    app = QApplication()
    x = DebugWidget().setup()
    x.add_show_func_result_button(function=stupid_function, category_name="stupid")
    x.add_show_func_result_button(function=stupid_function_2, category_name="stupid")
    x.add_show_func_result_button(function=stupid_function_3, category_name="stupid")

    x.add_show_func_result_button(function=QApplication.screens, category_name="Qt")
    x.add_show_func_result_button(function=stupid_function_4, category_name="stupid")

    x.add_show_attr_button("__dict__", app)
    x.add_show_attr_button("__dict__", QDockWidget)
    x.add_show_attr_button("argv", sys)
    x.add_show_attr_button("path", sys)
    x.add_show_attr_button("base_prefix", sys)
    x.add_show_attr_button("base_exec_prefix", sys)
    x.add_show_attr_button("builtin_module_names", sys)
    x.add_show_attr_button("exec_prefix", sys)
    x.add_show_attr_button("exc_info", sys)
    x.add_show_attr_button("flags", sys)
    x.add_show_func_result_button(get_all_lexers, "pygments")
    x.add_show_func_result_button(formatters.get_all_formatters, "pygments")
    x.add_show_func_result_button(filters.get_all_filters, "pygments")
    x.add_show_func_result_button(inspect_all_filter, "pygments")
    x.add_show_func_result_button(function=stupid_function_5, category_name="stupid")

    x.add_show_attr_button("cwd", Path())

    x.show()
    app.exec()
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.general_helper.conversion import ns_to_s, number_to_pretty

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def print_profile_hook(db, stmt: str, time_taken: int):

    time_taken = ns_to_s(time_taken)
    time_taken = number_to_pretty(time_taken) + " s"
    print(f"Executed {stmt!r} in {time_taken}", flush=True)
# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Mapping, Iterable
from pathlib import Path
from collections import ChainMap

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import UnconvertableTypusError
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
from gidapptools.gid_config.conversion.base_converters import ConfigValueConverter, get_standard_converter
from gidapptools.gid_config.conversion.converter_grammar import ConverterSpecData

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class ConversionTable:
    __slots__ = ("_standard_converters", "_extra_converters", "_resolved_converters")

    def __init__(self,
                 extra_converter: Iterable[type[ConfigValueConverter]] = None) -> None:
        self._standard_converters: tuple[type[ConfigValueConverter]] = tuple(get_standard_converter())
        self._extra_converters: tuple[type[ConfigValueConverter]] = tuple(extra_converter or [])
        self._resolved_converters: dict[str, ConfigValueConverter] = None

    @property
    def converters(self) -> Mapping[str, type[ConfigValueConverter]]:
        if self._resolved_converters is None:

            self._resolved_converters = ChainMap(*self._resolve_converters(self._extra_converters), *self._resolve_converters(self._standard_converters))

        return self._resolved_converters

    @property
    def available_value_types(self) -> set[str]:
        return set(self.converters.keys())

    def _resolve_converters(self, converters: tuple[ConfigValueConverter]) -> tuple[Mapping[str, ConfigValueConverter]]:
        converter_map = {}
        converter_alias_map = {}
        for converter in converters:
            converter_map[converter.value_typus] = converter
            converter_alias_map = converter_alias_map | {a: converter for a in converter.value_typus_aliases}

        return converter_map, converter_alias_map

    def get_converter(self, converter_data: ConverterSpecData) -> ConfigValueConverter:

        converter = self.converters.get(converter_data["typus"], MiscEnum.NOT_FOUND)

        if converter is MiscEnum.NOT_FOUND:
            raise UnconvertableTypusError(f"No Converter for typus {converter_data['typus']!r} found.")

        return converter(self, **converter_data["kw_arguments"])

    def add_extra_converter(self, converter: type[ConfigValueConverter]) -> None:
        self._extra_converters = tuple(self._extra_converters + (converter,))
        self._resolved_converters = None


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import inspect
import logging
import warnings
from enum import Enum
from typing import Any, Callable, Optional, Iterable
from pathlib import Path
from functools import partial

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def _add_new_logging_level(name: str, value: int) -> None:
    if (name in logging._nameToLevel and value in logging._levelToName) and (logging._nameToLevel.get(name) == value and logging._levelToName.get(value) == name):
        return
    if name not in logging._nameToLevel:
        if logging._levelToName.get(value) is not None:
            raise ValueError(f"Value {value!r} already used by a different LoggingLevel ({logging._levelToName(value)!r}).")
        logging.addLevelName(value, name)
    elif value not in logging._levelToName:
        if logging._nameToLevel.get(name) is not None:
            raise ValueError(f"Name {name!r} already used by a different LoggingLevel ({logging._nameToLevel(name)!r}).")
        logging.addLevelName(value, name)
    else:
        raise RuntimeError(f"something went wrong with checking the LogLevel {name=}, {value=}.")


def _check_if_all_levels_are_in_LoggingLevel() -> None:
    if any(name not in set(LoggingLevel._member_map_.keys()) for name in logging._nameToLevel):
        missing_member_names = set(logging._nameToLevel.keys()).difference(set(LoggingLevel._member_map_.keys()))
        missing_members = [(name, logging._nameToLevel.get(name)) for name in missing_member_names]
        missing_members = sorted(missing_members, key=lambda x: (x[1], x[0]))
        missing_members_string = '\n\t\t'.join(f"(name: {item[0]!r}, value: {item[1]!r})" for item in missing_members)
        msg = f"{LoggingLevel.__name__!r} is missing logging level members ->\n\t\t{missing_members_string}"
        warnings.warn_explicit(message=msg, category=Warning, filename=THIS_FILE_DIR.name, lineno=inspect.findsource(LoggingLevel)[1], module=__name__, module_globals=globals(), source=inspect.getmodule(__name__))


def _align_left(text: str, width: int = 0) -> str:
    return text.ljust(width)


def _align_center(text: str, width: int = 0) -> str:
    return text.center(width)


def _align_right(text: str, width: int = 0) -> str:
    return text.rjust(width)


def _align_left_extra_padded(text: str, width: int = 0, amount_extra_padding: int = 1) -> str:
    extra_padding = " " * amount_extra_padding
    return (extra_padding + text).ljust(width)


def _align_none(text: str, width: int = 0) -> str:
    return text


class LoggingSectionAlignment(Enum):
    NONE = (_align_none, ("none",))

    LEFT = (_align_left, ("<",))
    CENTER = (_align_center, ("^",))
    RIGHT = (_align_right, (">",))

    LEFT_EXTRA_PADDED_ONCE = (partial(_align_left_extra_padded, amount_extra_padding=1), ("<+1", "l_pad_1"))
    LEFT_EXTRA_PADDED_TWICE = (partial(_align_left_extra_padded, amount_extra_padding=2), ("<+2", "l_pad_2"))
    LEFT_EXTRA_PADDED_THRICE = (partial(_align_left_extra_padded, amount_extra_padding=3), ("<+3", "l_pad_3"))

    def __init__(self, align_func: Callable[[str, Optional[int]], str], aliases: Iterable = None) -> None:
        self.align_func = align_func
        self.aliases = set(aliases) if aliases else set()

    def align(self, text: str, width: int = 0) -> str:
        return self.align_func(text, width)

    @classmethod
    def _missing_(cls, value: object) -> Any:
        if isinstance(value, str):
            mod_value = value.casefold()
            for member in cls.__members__.values():
                if mod_value == member.name.casefold():
                    return member
                if mod_value in member.aliases:
                    return member
        return super()._missing_(value)


class LoggingLevel(int, Enum):
    NOTSET = 0
    TRACE = 5
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    WARN = WARNING
    FATAL = CRITICAL

    def __init__(self, level: int) -> None:
        self.level = level
        if not self.is_alias:
            _add_new_logging_level(self.name, self.level)

    @classmethod
    def _missing_(cls, value: object) -> Any:
        if isinstance(value, str):
            mod_value = value.casefold()
            for member_name, member_value in cls.__members__.items():
                if member_name.casefold() == mod_value or member_value == mod_value:
                    return cls(member_value)
                if isinstance(member_value, str) and member_value.casefold() == mod_value:
                    return cls(member_value)
        return super()._missing_(value)

    @ property
    def is_alias(self) -> bool:
        return self.name in {"WARN", "FATAL"}

    def __index__(self) -> int:
        return self.value

    def __str__(self) -> str:
        return self.name


_check_if_all_levels_are_in_LoggingLevel()


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import queue
import logging
from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar, Callable
from pathlib import Path
from threading import Lock
from contextlib import contextmanager

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
log = logging.getLogger(__name__)
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class AbstractThreadsafePool(ABC):
    __slots__ = ("_lock", "_max_size", "_prefill", "_objects", "_queue")

    def __init__(self, max_size: int = None, prefill: bool = False) -> None:
        self._lock = Lock()
        self._max_size: int = max_size or 10
        self._prefill = prefill
        self._objects: list[object] = []
        self._queue = queue.Queue(maxsize=self._max_size)
        if self._prefill is True:
            while True:
                created = self._create_if_possible()
                if created is False:
                    break

    @property
    def max_size(self) -> int:
        return self._max_size

    @property
    def amount_objects(self) -> int:
        return len(self._objects)

    @abstractmethod
    def _create_new_object(self) -> object:
        ...

    def _create_if_possible(self) -> bool:
        if self.amount_objects >= self.max_size:
            return False
        new_object = self._create_new_object()
        self._objects.append(new_object)
        self._queue.put_nowait(new_object)
        return True

    def _get_object(self) -> object:
        with self._lock:
            try:
                return self._queue.get_nowait()
            except queue.Empty:
                self._create_if_possible()
                return self._queue.get(block=True)

    @contextmanager
    def __call__(self) -> Any:
        obj = self._get_object()

        yield obj

        self._queue.put_nowait(obj)
        self._queue.task_done()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(max_size={self.max_size!r}, prefill={self._prefill!r})'


class GenericThreadsafePool(AbstractThreadsafePool):
    __slots__ = ("_obj_creator",)

    def __init__(self, obj_creator: Callable, max_size: int = 0, prefill: bool = False) -> None:
        self._obj_creator = obj_creator
        super().__init__(max_size=max_size, prefill=prefill)

    def _create_new_object(self) -> object:

        return self._obj_creator()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(obj_creator={self._obj_creator!r}, max_size={self.max_size!r}, prefill={self._prefill!r})'


T = TypeVar('T', Type, Callable)


class DecorateAbleList(list[T]):

    def __call__(self, item: T) -> T:
        super().append(item)
        return item


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import json
from typing import Any
from pathlib import Path

# * Gid Imports ----------------------------------------------------------------------------------------->

import importlib.resources as importlib_resources

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
JSON_DATA_DIR = importlib_resources.files("gidapptools.data.json")
WEBCOLORS_JSON = JSON_DATA_DIR.joinpath("webcolors.json")

# endregion [Constants]


def _load_webcolors_data() -> tuple[dict[str:tuple[int, int, int, float]]]:
    with WEBCOLORS_JSON.open('r', encoding='utf-8', errors='ignore') as f:
        data = json.load(f)
    _out = []
    for item_data in data:
        new_item_data = {}
        new_item_data["name"] = item_data.get("name").casefold()
        new_item_data["value"] = (item_data.get("rgb").get('r'), item_data.get("rgb").get('g'), item_data.get("rgb").get('b'), 1.0)
        _out.append(new_item_data)
    return tuple(_out)


ALL_WEBCOLORS: tuple[dict[str:tuple[int, int, int, float]]] = _load_webcolors_data()

ALL_WEBCOLORS_BY_NAME: dict[str, tuple[int, int, int, float]] = {i["name"]: i["value"] for i in ALL_WEBCOLORS}

ALL_WEBCOLORS_BY_VALUE: dict[tuple[int, int, int, float], str] = {i["value"]: i["name"] for i in ALL_WEBCOLORS}


# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]
